"""Class and functions for ``[vak.eval]`` table in configuration file."""

from __future__ import annotations

import pathlib

from attrs import converters, define, field, validators
from attrs.validators import instance_of

from ..common.converters import expanded_user_path
from .dataset import DatasetConfig
from .model import ModelConfig
from .trainer import TrainerConfig


def convert_post_tfm_kwargs(post_tfm_kwargs: dict) -> dict:
    post_tfm_kwargs = dict(post_tfm_kwargs)

    if "min_segment_dur" not in post_tfm_kwargs:
        # because there's no null in TOML,
        # users leave arg out of config then we set it to None
        post_tfm_kwargs["min_segment_dur"] = None
    else:
        post_tfm_kwargs["min_segment_dur"] = float(
            post_tfm_kwargs["min_segment_dur"]
        )

    if "majority_vote" not in post_tfm_kwargs:
        # set default for this one too
        post_tfm_kwargs["majority_vote"] = False
    else:
        post_tfm_kwargs["majority_vote"] = bool(
            post_tfm_kwargs["majority_vote"]
        )

    return post_tfm_kwargs


def are_valid_post_tfm_kwargs(instance, attribute, value):
    """check if ``post_tfm_kwargs`` is valid"""
    if not isinstance(value, dict):
        raise TypeError(
            "'post_tfm_kwargs' should be declared in toml config as an inline table "
            f"that parses as a dict, but type was: {type(value)}. "
            "Please declare in a similar fashion: `{majority_vote = True, min_segment_dur = 0.02}`"
        )
    if any(
        [k not in {"majority_vote", "min_segment_dur"} for k in value.keys()]
    ):
        invalid_kwargs = [
            k
            for k in value.keys()
            if k not in {"majority_vote", "min_segment_dur"}
        ]
        raise ValueError(
            f"Invalid keyword argument name specified for 'post_tfm_kwargs': {invalid_kwargs}."
            "Valid names are: {'majority_vote', 'min_segment_dur'}"
        )
    if "majority_vote" in value:
        if not isinstance(value["majority_vote"], bool):
            raise TypeError(
                "'post_tfm_kwargs' keyword argument 'majority_vote' "
                f"should be of type bool but was: {type(value['majority_vote'])}"
            )
    if "min_segment_dur" in value:
        if value["min_segment_dur"] and not isinstance(
            value["min_segment_dur"], float
        ):
            raise TypeError(
                "'post_tfm_kwargs' keyword argument 'min_segment_dur' type "
                f"should be float but was: {type(value['min_segment_dur'])}"
            )


REQUIRED_KEYS = (
    "checkpoint_path",
    "dataset",
    "output_dir",
    "model",
    "trainer",
)


@define
class EvalConfig:
    """Class that represents [vak.eval] table in configuration file.

    Attributes
    ----------
    checkpoint_path : str
        path to directory with checkpoint files saved by Torch, to reload model
    output_dir : str
        Path to location where .csv files with evaluation metrics should be saved.
    model : vak.config.ModelConfig
        The model to use: its name,
        and the parameters to configure it.
        Must be an instance of :class:`vak.config.ModelConfig`
    batch_size : int
        number of samples per batch presented to models during training.
    dataset : vak.config.DatasetConfig
        The dataset to use: the path to it,
        and optionally a path to a file representing splits,
        and the name, if it is a built-in dataset.
        Must be an instance of :class:`vak.config.DatasetConfig`.
    trainer : vak.config.TrainerConfig
        Configuration for :class:`lightning.pytorch.Trainer`.
        Must be an instance of :class:`vak.config.TrainerConfig`.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    labelmap_path : str
        path to 'labelmap.json' file.
    frames_standardizer_path : str
        path to a saved :class:`vak.transforms.FramesStandardizer` object used to standardize (normalize) frames.
        If spectrograms were normalized and this is not provided, will give
        incorrect results.
    post_tfm_kwargs : dict
        Keyword arguments to post-processing transform.
        If None, then no additional clean-up is applied
        when transforming labeled timebins to segments,
        the default behavior.
        The transform used is
        ``vak.transforms.frame_labels.PostProcess`.
        Valid keyword argument names are 'majority_vote'
        and 'min_segment_dur', and should be appropriate
        values for those arguments: Boolean for ``majority_vote``,
        a float value for ``min_segment_dur``.
        See the docstring of the transform for more details on
        these arguments and how they work.
    """

    # required, external files
    checkpoint_path: pathlib.Path = field(converter=expanded_user_path)
    output_dir: pathlib.Path = field(converter=expanded_user_path)

    # required, model / dataloader
    model = field(
        validator=instance_of(ModelConfig),
    )
    batch_size = field(converter=int, validator=instance_of(int))
    dataset: DatasetConfig = field(
        validator=instance_of(DatasetConfig),
    )
    trainer: TrainerConfig = field(
        validator=instance_of(TrainerConfig),
    )

    # "optional" but actually required for frame classification models
    # TODO: check model family in __post_init__ and raise ValueError if labelmap
    # TODO: not specified for a frame classification model?
    labelmap_path = field(
        converter=converters.optional(expanded_user_path), default=None
    )
    # optional, transform
    frames_standardizer_path = field(
        converter=converters.optional(expanded_user_path),
        default=None,
    )

    post_tfm_kwargs = field(
        validator=validators.optional(are_valid_post_tfm_kwargs),
        converter=converters.optional(convert_post_tfm_kwargs),
        default=None,
    )

    # optional, data loader
    num_workers = field(validator=instance_of(int), default=2)

    @classmethod
    def from_config_dict(cls, config_dict: dict) -> EvalConfig:
        """Return :class:`EvalConfig` instance from a :class:`dict`.

        The :class:`dict` passed in should be the one found
        by loading a valid configuration toml file with
        :func:`vak.config.parse.from_toml_path`,
        and then using key ``eval``,
        i.e., ``EvalConfig.from_config_dict(config_dict['eval'])``."""
        for required_key in REQUIRED_KEYS:
            if required_key not in config_dict:
                raise KeyError(
                    "The `[vak.eval]` table in a configuration file requires "
                    f"the option '{required_key}', but it was not found "
                    "when loading the configuration file into a Python dictionary. "
                    "Please check that the configuration file is formatted correctly."
                )
        config_dict["dataset"] = DatasetConfig.from_config_dict(
            config_dict["dataset"]
        )
        config_dict["model"] = ModelConfig.from_config_dict(
            config_dict["model"]
        )
        config_dict["trainer"] = TrainerConfig(**config_dict["trainer"])
        return cls(**config_dict)



"""Class that represents the TOML configuration file used with the vak command-line interface."""

from __future__ import annotations

import pathlib

from attr.validators import instance_of, optional
from attrs import define, field

from . import load
from .eval import EvalConfig
from .learncurve import LearncurveConfig
from .predict import PredictConfig
from .prep import PrepConfig
from .train import TrainConfig
from .validators import are_keys_valid, are_tables_valid

TABLE_CLASSES_MAP = {
    "eval": EvalConfig,
    "learncurve": LearncurveConfig,
    "predict": PredictConfig,
    "prep": PrepConfig,
    "train": TrainConfig,
}


def _validate_tables_to_parse_arg_convert_list(
    tables_to_parse: str | list[str],
) -> list[str]:
    """Helper function used by :func:`from_toml` that
    validates the ``tables_to_parse`` argument,
    and returns it as a list of strings."""
    if isinstance(tables_to_parse, str):
        tables_to_parse = [tables_to_parse]

    if not isinstance(tables_to_parse, list):
        raise TypeError(
            f"`tables_to_parse` should be a string or list of strings but type was: {type(tables_to_parse)}"
        )

    if not all(
        [isinstance(table_name, str) for table_name in tables_to_parse]
    ):
        raise ValueError(
            "All table names in 'tables_to_parse' should be strings"
        )
    if not all(
        [
            table_name in list(TABLE_CLASSES_MAP.keys())
            for table_name in tables_to_parse
        ]
    ):
        raise ValueError(
            "All table names in 'tables_to_parse' should be valid names of tables. "
            f"Values for 'tables were: {tables_to_parse}.\n"
            f"Valid table names are: {list(TABLE_CLASSES_MAP.keys())}"
        )
    return tables_to_parse


@define
class Config:
    """Class that represents the TOML configuration file used with the vak command-line interface.

    Attributes
    ----------
    prep : vak.config.prep.PrepConfig
        Represents ``[vak.prep]`` table of config.toml file
    train : vak.config.train.TrainConfig
        Represents ``[vak.train]`` table of config.toml file
    eval : vak.config.eval.EvalConfig
        Represents ``[vak.eval]`` table of config.toml file
    predict : vak.config.predict.PredictConfig
        Represents ``[vak.predict]`` table of config.toml file.
    learncurve : vak.config.learncurve.LearncurveConfig
        Represents ``[vak.learncurve]`` table of config.toml file
    """

    prep = field(validator=optional(instance_of(PrepConfig)), default=None)
    train = field(validator=optional(instance_of(TrainConfig)), default=None)
    eval = field(validator=optional(instance_of(EvalConfig)), default=None)
    predict = field(
        validator=optional(instance_of(PredictConfig)), default=None
    )
    learncurve = field(
        validator=optional(instance_of(LearncurveConfig)), default=None
    )

    @classmethod
    def from_config_dict(
        cls,
        config_dict: dict,
        tables_to_parse: str | list[str] | None = None,
        toml_path: str | pathlib.Path | None = None,
    ) -> "Config":
        """Return instance of :class:`Config` class,
        given a :class:`dict` containing the contents of
        a TOML configuration file.

        This :func:`classmethod` expects the output
        of :func:`vak.config.load._load_from_toml_path`,
        that converts a :class:`tomlkit.TOMLDocument`
        to a :class:`dict`, and returns the :class:`dict`
        that is accessed by the top-level key ``'vak'``.

        Parameters
        ----------
        config_dict : dict
            Python ``dict`` containing a .toml configuration file,
            parsed by the ``toml`` library.
        toml_path : str, pathlib.Path
            path to a configuration file in TOML format. Default is None.
            Not required, used only to make any error messages clearer.
        tables_to_parse : str, list
            Name of top-level table or tables from configuration
            file that should be parsed. Can be a string
            (single table) or list of strings (multiple
            tables). Default is None,
            in which case all are validated and parsed.

        Returns
        -------
        config : vak.config.parse.Config
            instance of :class:`Config` class,
            whose attributes correspond to the
            top-level tables in a config.toml file.
        """
        are_tables_valid(config_dict, toml_path)
        if tables_to_parse is None:
            tables_to_parse = list(
                config_dict.keys()
            )  # i.e., parse all top-level tables
        else:
            tables_to_parse = _validate_tables_to_parse_arg_convert_list(
                tables_to_parse
            )

        config_kwargs = {}
        for table_name in tables_to_parse:
            if table_name in config_dict:
                are_keys_valid(config_dict, table_name, toml_path)
                table_config_dict = config_dict[table_name]
                config_kwargs[table_name] = TABLE_CLASSES_MAP[
                    table_name
                ].from_config_dict(table_config_dict)
            else:
                raise KeyError(
                    f"A table specified in `tables_to_parse` was not found in the config: {table_name}"
                )

        return cls(**config_kwargs)

    @classmethod
    def from_toml_path(
        cls,
        toml_path: str | pathlib.Path,
        tables_to_parse: list[str] | None = None,
    ) -> "Config":
        """Return instance of :class:`Config` class,
        given the path to a TOML configuration file.

        Parameters
        ----------
        toml_path : str, pathlib.Path
            Path to a configuration file in TOML format.
            Parsed by ``toml`` library, then converted to an
            instance of ``vak.config.parse.Config`` by
            calling ``vak.parse.from_toml``
        tables_to_parse : str, list
            Name of table or tables from configuration
            file that should be parsed. Can be a string
            (single table) or list of strings (multiple
            tables). Default is None,
            in which case all are validated and parsed.

        Returns
        -------
        config : vak.config.parse.Config
            instance of :class:`Config` class, whose attributes correspond to
            tables in a config.toml file.
        """
        config_dict: dict = load._load_toml_from_path(toml_path)
        return cls.from_config_dict(config_dict, tables_to_parse, toml_path)



"""A LightningModule that represents a task
where a model predicts a label for each frame
in a time series, e.g., each time bin in
a window from a spectrogram."""

from __future__ import annotations

import logging
from typing import Callable, Mapping

import lightning
import torch

from .. import common, transforms
from ..common import labels
from .registry import model_family

logger = logging.getLogger(__name__)


@model_family
class FrameClassificationModel(lightning.LightningModule):
    """Class that represents a family of neural network models
    that predicts a label for each frame in a time series,
    e.g., each time bin in a window from a spectrogram.

    The task of predicting a label for each frame in a series
    is one way of predicting annotations for a vocalization,
    where the annotations consist of a sequence
    of segments, each with an onset, offset, and label.
    The model maps the spectrogram window
    to a vector of labels for each frame, i.e., each time bin.

    To annotate a vocalization with such a model,
    the spectrogram is converted into a batch of
    consecutive non-overlapping windows,
    for which the model produces predictions.
    These predictions are then concatenated
    into a vector of labeled frames,
    from which the segments can be recovered.

    Post-processing can be applied to the vector
    to clean up noisy predictions
    before recovering the segments.

    Attributes
    ----------
    network : torch.nn.Module, dict
        An instance of a ``torch.nn.Module``
        that implements a neural network,
        or a ``dict`` that maps human-readable string names
        to a set of such instances.
    loss : torch.nn.Module, callable
        An instance of a ``torch.nn.Module`    `
        that implements a loss function,
        or a callable Python function that
        computes a scalar loss.
    optimizer : torch.optim.Optimizer
        An instance of a ``torch.optim.Optimizer`` class
        used with ``loss`` to optimize
        the parameters of ``network``.
    metrics : dict
        A ``dict`` that maps human-readable string names
        to ``Callable`` functions, used to measure
        performance of the model.
    post_tfm : callable
        Post-processing transform applied to predictions.
    labelmap : dict-like
        That maps human-readable labels to integers predicted by network.
    eval_labelmap : dict-like
        Mapping from labels to integers predicted by network
        that is used by ``validation_step``.
        This is used when mapping from network outputs back to labels
        to compute metrics that require strings, such as edit distance.
        If ``labelmap`` contains keys with multiple characters,
        this will be ``labelmap`` re-mapped so that all labels have
        single characters (except the background label, if specified),
        to avoid artificially changing the edit distance.
        See https://github.com/vocalpy/vak/issues/373 for more detail.
        If all keys (except background label) are single-character,
        then ``eval_labelmap`` will just be ``labelmap``.
    to_labels_eval : vak.transforms.frame_labels.ToLabels
        Instance of :class:`~vak.transforms.frame_labels.ToLabels`
        that uses ``eval_labelmap`` to convert labeled timebins
        to string labels inside of ``validation_step``,
        for computing edit distance.
    """

    def __init__(
        self,
        labelmap: Mapping,
        network: torch.nn.Module | dict | None = None,
        loss: torch.nn.Module | Callable | None = None,
        optimizer: torch.optim.Optimizer | None = None,
        metrics: dict | None = None,
        post_tfm: Callable | None = None,
        background_label=common.constants.DEFAULT_BACKGROUND_LABEL,
    ):
        """Initialize a new instance of a
        :class:`~vak.models.frame_classification_model.FrameClassificationModel`.

        Parameters
        ----------
        labelmap : dict-like
            That maps human-readable labels to integers predicted by network.
        network : torch.nn.Module, dict
            An instance of a :class:`torch.nn.Module`
            that implements a neural network,
            or a ``dict`` that maps human-readable string names
            to a set of such instances.
        loss : torch.nn.Module, callable
            An instance of a :class:`torch.nn.Module`
            that implements a loss function,
            or a callable Python function that
            computes a scalar loss.
        optimizer : torch.optim.Optimizer
            An instance of a ``torch.optim.Optimizer`` class
            used with ``loss`` to optimize
            the parameters of ``network``.
        metrics : dict
            A ``dict`` that maps human-readable string names
            to ``Callable`` functions, used to measure
            performance of the model.
        post_tfm : callable
            Post-processing transform applied to predictions.
        background_label: str, optional
            The string label applied to segments belonging to the
            background class.
            Default is
            :const:`vak.common.constants.DEFAULT_BACKGROUND_LABEL`.
        """
        super().__init__()
        self.network = network
        self.loss = loss
        self.optimizer = optimizer
        self.metrics = metrics

        self.labelmap = labelmap
        # replace any multiple character labels in mapping
        # with single-character labels
        # so that we do not affect edit distance computation
        # see https://github.com/NickleDave/vak/issues/373
        labelmap_keys = [
            lbl for lbl in labelmap.keys() if lbl != background_label
        ]
        if any(
            [len(label) > 1 for label in labelmap_keys]
        ):  # only re-map if necessary
            # (to minimize chance of knock-on bugs)
            logger.info(
                "Detected that labelmap has keys with multiple characters:"
                f"\n{labelmap_keys}\n"
                "Re-mapping labelmap used with to_labels_eval transform, using "
                "function vak.labels.multi_char_labels_to_single_char"
            )
            self.eval_labelmap = labels.multi_char_labels_to_single_char(
                labelmap
            )
        else:
            self.eval_labelmap = labelmap

        self.to_labels_eval = transforms.frame_labels.ToLabels(
            self.eval_labelmap
        )
        self.post_tfm = post_tfm

    def configure_optimizers(self):
        """Returns the model's optimizer.

        Method required by ``lightning.pytorch.LightningModule``.
        This method returns the ``optimizer`` instance passed into ``__init__``.
        If None was passed in, an instance that was created
        with default arguments will be returned.
        """
        return self.optimizer

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run a forward pass through this model's network.

        Parameters
        ----------
        x : torch.Tensor
            Input to network, with shape that matches ``self.input_shape``.

        Returns
        -------
        out : torch.Tensor
            Output from network.
        """
        return self.network(x)

    def training_step(self, batch: tuple, batch_idx: int):
        """Perform one training step.

        Method required by ``lightning.pytorch.LightningModule``.

        Parameters
        ----------
        batch : tuple
            A batch from a dataloader.
        batch_idx : int
            The index of this batch in the dataloader.

        Returns
        -------
        loss : torch.Tensor
            Scalar loss value computed by
            the loss function, ``self.loss``.
        """
        frames = batch["frames"]

        # we repeat this code in validation step
        # because I'm assuming it's faster than a call to a staticmethod that factors it out
        if (  # multi-class frame classificaton
            "multi_frame_labels" in batch
            and "binary_frame_labels" not in batch
            and "boundary_frame_labels" not in batch
        ):
            target_types = ("multi_frame_labels",)
        elif (  # binary frame classification
            "binary_frame_labels" in batch
            and "multi_frame_labels" not in batch
            and "boundary_frame_labels" not in batch
        ):
            target_types = ("binary_frame_labels",)
        elif (  # boundary "detection" -- i.e. different kind of binary frame classification
            "boundary_frame_labels" in batch
            and "multi_frame_labels" not in batch
            and "binary_frame_labels" not in batch
        ):
            target_types = ("boundary_frame_labels",)
        elif (  # multi-class frame classification *and* boundary detection
            "multi_frame_labels" in batch
            and "boundary_frame_labels" in batch
            and "binary_frame_labels" not in batch
        ):
            target_types = ("multi_frame_labels", "boundary_frame_labels")

        if len(target_types) == 1:
            class_logits = self.network(frames)
            loss = self.loss(class_logits, batch[target_types[0]])
            self.log("train_loss", loss, on_step=True)
        else:
            multi_logits, boundary_logits = self.network(frames)
            loss = self.loss(
                multi_logits,
                boundary_logits,
                batch["multi_frame_labels"],
                batch["boundary_frame_labels"],
            )
            if isinstance(loss, torch.Tensor):
                self.log("train_loss", loss, on_step=True)
            elif isinstance(loss, dict):
                # this provides a mechanism to values for all terms of a loss function with multiple terms
                for loss_name, loss_val in loss.items():
                    self.log(f"train_{loss_name}", loss_val, on_step=True)

        return loss

    def validation_step(self, batch: tuple, batch_idx: int):
        """Perform one validation step.

        Method required by ``lightning.pytorch.LightningModule``.
        Logs metrics using ``self.log``

        Parameters
        ----------
        batch : tuple
            A batch from a dataloader.
        batch_idx : int
            The index of this batch in the dataloader.

        Returns
        -------
        None
        """
        frames = batch["frames"]
        # remove "batch" dimension added by collate_fn to frames
        # TODO: fix this weirdness. Diff't collate_fn?
        if frames.ndim in (5, 4):
            if frames.shape[0] == 1:
                frames = torch.squeeze(frames, dim=0)
        else:
            raise ValueError(f"invalid shape for frames: {frames.shape}")

        # we repeat this code in training step
        # because I'm assuming it's faster than a call to a staticmethod that factors it out
        if (  # multi-class frame classificaton
            "multi_frame_labels" in batch
            and "binary_frame_labels" not in batch
            and "boundary_frame_labels" not in batch
        ):
            target_types = ("multi_frame_labels",)
        elif (  # binary frame classification
            "binary_frame_labels" in batch
            and "multi_frame_labels" not in batch
            and "boundary_frame_labels" not in batch
        ):
            target_types = ("binary_frame_labels",)
        elif (  # boundary "detection" -- i.e. different kind of binary frame classification
            "boundary_frame_labels" in batch
            and "multi_frame_labels" not in batch
            and "binary_frame_labels" not in batch
        ):
            target_types = ("boundary_frame_labels",)
        elif (  # multi-class frame classification *and* boundary detection
            "multi_frame_labels" in batch
            and "boundary_frame_labels" in batch
            and "binary_frame_labels" not in batch
        ):
            target_types = ("multi_frame_labels", "boundary_frame_labels")

        if len(target_types) == 1:
            class_logits = self.network(frames)
            boundary_logits = None
        else:
            class_logits, boundary_logits = self.network(frames)

        # permute and flatten out
        # so that it has shape (1, number classes, number of time bins)
        # ** NOTICE ** just calling out.reshape(1, out.shape(1), -1) does not work, it will change the data
        class_logits = class_logits.permute(1, 0, 2)
        class_logits = torch.flatten(class_logits, start_dim=1)
        class_logits = torch.unsqueeze(class_logits, dim=0)
        # reduce to predictions, assuming class dimension is 1
        class_preds = torch.argmax(
            class_logits, dim=1
        )  # y_pred has dims (batch size 1, predicted label per time bin)

        if boundary_logits is not None:
            boundary_logits = boundary_logits.permute(1, 0, 2)
            boundary_logits = torch.flatten(boundary_logits, start_dim=1)
            boundary_logits = torch.unsqueeze(boundary_logits, dim=0)
            # reduce to predictions, assuming class dimension is 1
            boundary_preds = torch.argmax(
                boundary_logits, dim=1
            )  # y_pred has dims (batch size 1, predicted label per time bin)

        if "padding_mask" in batch:
            padding_mask = batch[
                "padding_mask"
            ]  # boolean: 1 where valid, 0 where padding
            # remove "batch" dimension added by collate_fn
            # because this extra dimension just makes it confusing to use the mask as indices
            if padding_mask.ndim == 2:
                if padding_mask.shape[0] == 1:
                    padding_mask = torch.squeeze(padding_mask, dim=0)
            else:
                raise ValueError(
                    f"invalid shape for padding mask: {padding_mask.shape}"
                )

            class_logits = class_logits[:, :, padding_mask]
            class_preds = class_preds[:, padding_mask]

            if boundary_logits is not None:
                boundary_logits = boundary_logits[:, :, padding_mask]
                boundary_preds = boundary_preds[:, padding_mask]

        if "multi_frame_labels" in target_types:
            multi_frame_labels_str = self.to_labels_eval(
                batch["multi_frame_labels"].cpu().numpy()
            )
            class_preds_str = self.to_labels_eval(class_preds.cpu().numpy())

            if self.post_tfm:
                if target_types == ("multi_frame_labels",):
                    class_preds_tfm = self.post_tfm(
                        class_preds.cpu().numpy(),
                    )
                elif target_types == ("multi_frame_labels", "boundary_frame_labels"):
                    class_preds_tfm = self.post_tfm(
                        class_preds.cpu().numpy(),
                        boundary_labels=boundary_preds.cpu().numpy(),
                    )
                class_preds_tfm_str = self.to_labels_eval(class_preds_tfm)
                # convert back to tensor so we can compute accuracy
                class_preds_tfm = torch.from_numpy(class_preds_tfm).to(
                    self.device
                )

        if len(target_types) == 1:
            target = batch[target_types[0]]
        else:
            target = {
                target_type: batch[target_type] for target_type in target_types
            }

        for metric_name, metric_callable in self.metrics.items():
            if metric_name == "loss":
                if len(target_types) == 1:
                    self.log(
                        f"val_{metric_name}",
                        metric_callable(class_logits, target),
                        batch_size=1,
                        on_step=True,
                        sync_dist=True,
                    )
                else:
                    loss = self.loss(
                        class_logits,
                        boundary_logits,
                        target["multi_frame_labels"],
                        target["boundary_frame_labels"],
                    )
                    if isinstance(loss, torch.Tensor):
                        self.log(
                            f"val_{metric_name}",
                            loss,
                            batch_size=1,
                            on_step=True,
                            sync_dist=True,
                        )
                    elif isinstance(loss, dict):
                        # this provides a mechanism to values for all terms of a loss function with multiple terms
                        for loss_name, loss_val in loss.items():
                            self.log(
                                f"val_{loss_name}",
                                loss_val,
                                batch_size=1,
                                on_step=True,
                                sync_dist=True,
                            )
            elif metric_name == "acc":
                if len(target_types) == 1:
                    self.log(
                        f"val_{metric_name}",
                        metric_callable(class_preds, target),
                        batch_size=1,
                        on_step=True,
                        sync_dist=True,
                    )
                    if self.post_tfm and "multi_frame_labels" in target_types:
                        self.log(
                            f"val_{metric_name}_tfm",
                            metric_callable(class_preds_tfm, target),
                            batch_size=1,
                            on_step=True,
                            sync_dist=True,
                        )
                else:
                    self.log(
                        f"val_multi_{metric_name}",
                        metric_callable(
                            class_preds, target["multi_frame_labels"]
                        ),
                        batch_size=1,
                        on_step=True,
                        sync_dist=True,
                    )
                    self.log(
                        f"val_boundary_{metric_name}",
                        metric_callable(
                            boundary_preds, target["boundary_frame_labels"]
                        ),
                        batch_size=1,
                        on_step=True,
                        sync_dist=True,
                    )
                    if self.post_tfm and "multi_frame_labels" in target_types:
                        self.log(
                            f"val_multi_{metric_name}_tfm",
                            metric_callable(
                                class_preds_tfm, target["multi_frame_labels"]
                            ),
                            batch_size=1,
                            on_step=True,
                            sync_dist=True,
                        )
            elif (
                metric_name == "levenshtein"
                or metric_name == "character_error_rate"
            ) and "multi_frame_labels" in target_types:
                self.log(
                    f"val_{metric_name}",
                    # next line: convert to float to squelch warning from lightning
                    float(
                        metric_callable(
                            class_preds_str, multi_frame_labels_str
                        )
                    ),
                    batch_size=1,
                    on_step=True,
                    sync_dist=True,
                )
                if self.post_tfm:
                    self.log(
                        f"val_{metric_name}_tfm",
                        # next line: convert to float to squelch warning from lightning
                        float(
                            metric_callable(
                                class_preds_tfm_str, multi_frame_labels_str
                            )
                        ),
                        batch_size=1,
                        on_step=True,
                        sync_dist=True,
                    )

    def predict_step(self, batch: tuple, batch_idx: int):
        """Perform one prediction step.

        Method required by ``lightning.pytorch.LightningModule``.

        Parameters
        ----------
        batch : tuple
            A batch from a dataloader.
        batch_idx : int
            The index of this batch in the dataloader.

        Returns
        -------
        y_pred : dict
            Where the key is "source_path" and the value
            is the output of the network;
            "source_path" is the path to the file
            containing the spectrogram
            for which a prediction was generated.
        """
        frames, frames_path = (
            batch["frames"].to(self.device),
            batch["frames_path"],
        )
        if isinstance(frames_path, list) and len(frames_path) == 1:
            frames_path = frames_path[0]
        # TODO: fix this weirdness. Diff't collate_fn?
        if frames.ndim in (5, 4):
            if frames.shape[0] == 1:
                frames = torch.squeeze(frames, dim=0)
        else:
            raise ValueError(f"invalid shape for `frames`: {frames.shape}")
        y_pred = self.network(frames)
        return {frames_path: y_pred}

    def load_state_dict_from_path(self, ckpt_path):
        """Loads a model from the path to a saved checkpoint.

        Loads the checkpoint and then calls
        ``self.load_state_dict`` with the ``state_dict``
        in that chekcpoint.

        This method allows loading a state dict into an instance.
        It's necessary because `lightning.pytorch.LightningModule.load`` is a
        ``classmethod``, so calling that method will trigger
         ``LightningModule.__init__`` instead of running
        ``vak.models.Model.__init__``.

        Parameters
        ----------
        ckpt_path : str, pathlib.Path
            Path to a checkpoint saved by a model in ``vak``.
            This checkpoint has the same key-value pairs as
            any other checkpoint saved by a
            ``lightning.pytorch.LightningModule``.

        Returns
        -------
        None

        This method modifies the model state by loading the ``state_dict``;
        it does not return anything.
        """
        ckpt = torch.load(ckpt_path)
        self.load_state_dict(ckpt["state_dict"])



"""Constants used by vaktestdata and scripts that rely on it."""
import json
import pathlib

from .config_metadata import ConfigMetadata

HERE = pathlib.Path(__file__).parent
TEST_DATA_ROOT = HERE / ".." / ".." / "data_for_tests"
CONFIG_METADATA_JSON_PATH = TEST_DATA_ROOT / "configs" / "configs.json"
with CONFIG_METADATA_JSON_PATH.open('r') as fp:
    CONFIG_METADATA_LIST = json.load(fp)['config_metadata']
CONFIG_METADATA = [
    ConfigMetadata(**config_metadata_dict)
    for config_metadata_dict in CONFIG_METADATA_LIST
]
GENERATED_TEST_DATA_ROOT = TEST_DATA_ROOT / "generated"

GENERATED_SPECT_OUTPUT_DIR = GENERATED_TEST_DATA_ROOT / "spect-output-dir"
GENERATED_SOURCE_FILES_CSV_DIR = GENERATED_TEST_DATA_ROOT / "source-files-csv"
GENERATED_SOURCE_FILES_WITH_SPLITS_CSV_DIR = GENERATED_TEST_DATA_ROOT / "source-files-with-splits-csv"

GENERATED_TEST_CONFIGS_ROOT = GENERATED_TEST_DATA_ROOT / "configs"

# convention is that all the config.toml files in tests/data_for_tests/configs
# that should be run when generating test data
# have filenames of the form `{MODEL}_{COMMAND}_audio_{FORMAT}_annot_{FORMAT}.toml'
# **or** `{MODEL}_{COMMAND}_spect_{FORMAT}_annot_{FORMAT}_config.ini'
# e.g., 'TweetyNet_learncurve_audio_cbin_annot_notmat.toml'.
# Below, we iterate over model names
# so glob doesn't pick up static configs that are just used for testing,
# like 'invalid_option_config.toml`
TEST_CONFIGS_ROOT = TEST_DATA_ROOT.joinpath("configs")

# the sub-directories that will get made inside `./tests/data_for_tests/generated`
TOP_LEVEL_DIRS = [
    "prep",
    "results",
]

# need to run 'train' config before we run 'predict'
# so we can add checkpoints, etc., from training to predict
COMMANDS = (
    "train",
    "learncurve",
    "eval",
    "predict",
    "train_continue",
)

GENERATE_TEST_DATA_STEPS = (
    'prep',
    'results',
    'all',
)



import copy

import pytest

import torch
from vak import metrics
from vak.nets import TweetyNet

import vak.models

from .conftest import (
    MockEncoder,
    MockDecoder,
)


class TweetyNetDefinition:
    """Redefine here to test that ``vak.models.definition.validate``
    actually works on classes we care about.

    Can't use the class itself in the ``vak.models.tweetynet``
    because that's already decorated with ``vak.models.decorator.model``.
    """
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}
    default_config = {
        'optimizer':
            {'lr': 0.003}
    }


class MissingClassVarModelDefinition:
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    # no metrics, missing class var


class ExtraClassVarModelDefinition:
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}
    extra_class_variable = TweetyNet


class InvalidNetworkTypeModelDefinition:
    network = torch.optim.Adam  # should be torch.nn.Module
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}


class InvalidNetworkDictKeyModelDefinition:
    network = {torch.optim.Adam: torch.optim.Adam}  # key should be a string
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}


class InvalidNetworkDictValueModelDefinition:
    network = {'adamnet': None}  # value should be torch.nn.Module
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}


class InvalidLossTypeModelDefinition:
    network = TweetyNet
    loss = None  # should be torch.nn.Module or callable
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}


class InvalidOptimTypeModelDefinition:
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = TweetyNet  # should be a torch.optim subclass
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}


class InvalidMetricsTypeModelDefinition:
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = metrics.Accuracy  # metrics should be a dict


class InvalidMetricsDictKeyModelDefinition:
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {metrics.Accuracy: metrics.Accuracy}  # key should  be str


class InvalidMetricsDictValueModelDefinition:
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': None}  # should be callable


class InvalidDefaultConfigNetworkKwargDefinition:
    """Definition with invalid keyword arg in 'network' config"""
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}
    default_config = {
        # invalid keyword arg should raise error
        'network': {'invalid_keyword_arg': 100000}
    }


class InvalidDefaultConfigNetworkDictKeyDefinition:
    """Definition that specifies 'network' as a dict
    but the default_config' has a key that is not any of those network names"""
    network = {'MockEncoder': MockEncoder, 'MockDecoder': MockDecoder}
    loss = torch.nn.TripletMarginWithDistanceLoss
    optimizer = torch.optim.Adam
    metrics = {
        'loss': torch.nn.TripletMarginWithDistanceLoss
    }
    default_config = {
        # invalid key that should be network name
        'invalid_network_name': {}
    }


class InvalidDefaultConfigNetworkDictKwargDefinition:
    """Definition that specifies 'network' as a dict
    but the default_config' has a kwarg that is not valid
    for one of the networks"""
    network = {'MockEncoder': MockEncoder, 'MockDecoder': MockDecoder}
    loss = torch.nn.TripletMarginWithDistanceLoss
    optimizer = torch.optim.Adam
    metrics = {
        'loss': torch.nn.TripletMarginWithDistanceLoss
    }
    default_config = {
        # invalid key that should be network name
        'MockEncoder': {'invalid_kwarg'},
        'MockDecoder': {},
    }


class InvalidDefaultConfigLossIsFunctionButKwargsModelDefinition:
    """Definition with default config that specifies kwargs for loss,
    but loss is a function (so it can't take kwargs on __init__,
    there's no __init__)."""
    network = TweetyNet
    loss = lambda y, y_pred: y == y_pred
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}
    default_config = {
        'loss': {'invalid_kwarg': 10000}
    }


class InvalidDefaultConfigLossKwarg:
    """Definition with invalid keyword arg in 'loss' config"""
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}
    default_config = {
        # invalid keyword arg should raise error
        'loss': {'invalid_keyword_arg': 100000}
    }


class InvalidDefaultConfigOptimizerKwarg:
    """Definition with invalid keyword arg in 'optimizer' config"""
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}
    default_config = {
        # invalid keyword arg should raise error
        'optimizer': {'invalid_keyword_arg': 100000}
    }


class InvalidDefaultConfigMetricName:
    """Definition with invalid metric name in 'metric' config"""
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}
    default_config = {
        # invalid metric name should raise error
        'metrics': {'invalid_metric_name': {}}
    }


class InvalidDefaultConfigMetricKwarg:
    """Definition with invalid metric name in 'metric' config"""
    network = TweetyNet
    loss = torch.nn.CrossEntropyLoss
    optimizer = torch.optim.Adam
    metrics = {'acc': metrics.Accuracy,
               'levenshtein': metrics.Levenshtein,
               'character_error_rate': metrics.CharacterErrorRate,
               'loss': torch.nn.CrossEntropyLoss}
    default_config = {
        # invalid keyword arg should raise error
        'metrics': {'acc': {'invalid_kwarg': 10000}}
    }


class TestModelDefinition:
    """Test the class ``vak.models.definition.ModelDefinition``"""

    @pytest.mark.parametrize(
        'definition',
        [
            TweetyNetDefinition,
        ]
    )
    def test_validate(self, definition):
        """Test that ``ModelDefinition.validate`` works as expected.

        This is the main method we need to test on this class,
        since it's used by other functions,
        e.g. ``vak.models.make.make`` to validate a class
        passed in that claims to be a model definition.
        """
        before = copy.deepcopy(definition)
        after = vak.models.definition.validate(definition)
        assert after is definition

        if not hasattr(before, 'default_config'):
            assert hasattr(after, 'default_config')
            assert isinstance(after.default_config, dict)
            assert after.default_config == vak.models.definition.DEFAULT_DEFAULT_CONFIG

        elif hasattr(before, 'default_config'):
            for key in vak.models.definition.DEFAULT_DEFAULT_CONFIG:
                if key not in before.default_config:
                    assert after.default_config[key] == {}
                else:
                    assert after.default_config[key] == before.default_config[key]

    @pytest.mark.parametrize(
        'definition, expected_error',
        [
            (MissingClassVarModelDefinition, ValueError),
            (ExtraClassVarModelDefinition, ValueError),
            (InvalidNetworkTypeModelDefinition, TypeError),
            (InvalidNetworkDictKeyModelDefinition, TypeError),
            (InvalidNetworkDictValueModelDefinition, TypeError),
            (InvalidLossTypeModelDefinition, TypeError),
            (InvalidOptimTypeModelDefinition, TypeError),
            (InvalidMetricsTypeModelDefinition, TypeError),
            (InvalidMetricsDictKeyModelDefinition, TypeError),
            (InvalidMetricsDictValueModelDefinition, TypeError),
            (InvalidDefaultConfigNetworkKwargDefinition, ValueError),
            (InvalidDefaultConfigNetworkDictKeyDefinition, ValueError),
            (InvalidDefaultConfigNetworkDictKwargDefinition, ValueError),
            (InvalidDefaultConfigLossIsFunctionButKwargsModelDefinition, ValueError),
            (InvalidDefaultConfigLossKwarg, ValueError),
            (InvalidDefaultConfigOptimizerKwarg, ValueError),
            (InvalidDefaultConfigMetricName, ValueError),
            (InvalidDefaultConfigMetricKwarg, ValueError),
        ]
    )
    def test_validate_raises(self, definition, expected_error):
        with pytest.raises(expected_error):
            vak.models.definition.validate(definition)



"""Tests for vak.eval.eval function."""
from unittest import mock

import pytest

import vak.config
import vak.common.constants
import vak.common.paths
import vak.eval


@pytest.mark.parametrize(
    "audio_format, spect_format, annot_format, model_name, eval_function_to_mock",
    [
        ("cbin", None, "notmat", "TweetyNet",
         'vak.eval.eval_.eval_frame_classification_model'),
        ("cbin", None, "notmat", "ConvEncoderUMAP",
         'vak.eval.eval_.eval_parametric_umap_model'),
    ],
)
def test_eval(
        audio_format, spect_format, annot_format, model_name, eval_function_to_mock,
        specific_config_toml_path, tmp_path
):
    """Test that :func:`vak.eval.eval` dispatches to the correct model-specific
    training functions"""
    output_dir = tmp_path.joinpath(
        f"test_eval_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    keys_to_change = [
        {"table": "eval", "key": "output_dir", "value": str(output_dir)},
        {"table": "eval", "key": "trainer", "value": {"accelerator": "cpu", "devices": 1}},
    ]

    toml_path = specific_config_toml_path(
        config_type="eval",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        keys_to_change=keys_to_change,
    )
    cfg = vak.config.Config.from_toml_path(toml_path)

    results_path = tmp_path / 'results_path'
    results_path.mkdir()

    with mock.patch(eval_function_to_mock, autospec=True) as mock_eval_function:
        vak.eval.eval(
            model_config=cfg.eval.model.asdict(),
            dataset_config=cfg.eval.dataset.asdict(),
            trainer_config=cfg.eval.trainer.asdict(),
            checkpoint_path=cfg.eval.checkpoint_path,
            labelmap_path=cfg.eval.labelmap_path,
            output_dir=cfg.eval.output_dir,
            num_workers=cfg.eval.num_workers,
            batch_size=cfg.eval.batch_size,
            frames_standardizer_path=cfg.eval.frames_standardizer_path,
            post_tfm_kwargs=cfg.eval.post_tfm_kwargs,
        )

        assert mock_eval_function.called



"""Parametric UMAP model, as described in [1]_, with a convolutional network as the encoder.

Code adapted from implementation by @elyxlz
https://github.com/elyxlz/umap_pytorch
with changes made by Tim Sainburg:
https://github.com/lmcinnes/umap/issues/580#issuecomment-1368649550.
"""

from __future__ import annotations

import torch

from .. import metrics, nets, nn
from .decorator import model
from .parametric_umap_model import ParametricUMAPModel


@model(family=ParametricUMAPModel)
class ConvEncoderUMAP:
    """Parametric UMAP model, as described in [1]_,
    with a convolutional network as the encoder.

    Attributes
    ----------
    network : dict
        A dict with two keys, 'encoder' and 'decoder'.
        The 'encoder is vak.nets.ConvEncoder,
        an encoder with convolutional layers.
        The 'decoder' defaults to None.
    loss: torch.nn.CrossEntropyLoss
        Standard cross-entropy loss
    optimizer: torch.optim.Adam
        Adam optimizer.
    metrics: dict
        Mapping string names to the following metrics:
        ``vak.metrics.Accuracy``, ``vak.metrics.Levenshtein``,
        ``vak.metrics.CharacterErrorRate``, ``torch.nn.CrossEntropyLoss``.

    Notes
    -----
    Code adapted from implementation by @elyxlz
    https://github.com/elyxlz/umap_pytorch
    with changes made by Tim Sainburg:
    https://github.com/lmcinnes/umap/issues/580#issuecomment-1368649550.

    References
    ----------
    .. [1] Sainburg, T., McInnes, L., & Gentner, T. Q. (2021).
       Parametric UMAP embeddings for representation and semisupervised learning.
       Neural Computation, 33(11), 2881-2907.
       https://direct.mit.edu/neco/article/33/11/2881/107068.

    """

    network = {"encoder": nets.ConvEncoder}
    loss = nn.UmapLoss
    optimizer = torch.optim.AdamW
    metrics = {
        "acc": metrics.Accuracy,
        "levenshtein": metrics.Levenshtein,
        "character_error_rate": metrics.CharacterErrorRate,
        "loss": torch.nn.CrossEntropyLoss,
    }
    default_config = {
        "optimizer": {"lr": 1e-3},
    }



import torch


def get_default() -> str:
    """Get default `accelerator` for :class:`lightning.pytorch.Trainer`.

    Returns
    -------
    accelerator : str
        Will be ``'gpu'`` if :func:`torch.cuda.is_available`
        is ``True``, and ``'cpu'`` if not.
    """
    if torch.cuda.is_available():
        return "gpu"
    else:
        return "cpu"



import pathlib

import pytest

import vak.config.dataset


class TestDatasetConfig:
    @pytest.mark.parametrize(
        'path, splits_path, name',
        [
            # typical use by a user with default split
            ('~/user/prepped/dataset', None, None),
            # use by a user with a split specified
            ('~/user/prepped/dataset', 'spilts/replicate-1.json', None),
            # use of a built-in dataset, with a split specified
            ('~/datasets/BioSoundSegBench', 'splits/Bengalese-Finch-song-gy6or6-replicate-1.json', 'BioSoundSegBench'),

        ]
    )
    def test_init(self, path, splits_path, name):
        if name is None and splits_path is None:
            dataset_config = vak.config.dataset.DatasetConfig(
                path=path
            )
        elif name is None:
            dataset_config = vak.config.dataset.DatasetConfig(
                path=path,
                splits_path=splits_path,
            )
        else:
            dataset_config = vak.config.dataset.DatasetConfig(
                name=name,
                path=path,
                splits_path=splits_path,
            )
        assert isinstance(dataset_config, vak.config.dataset.DatasetConfig)
        assert dataset_config.path == vak.common.converters.expanded_user_path(path)
        if splits_path is not None:
            assert dataset_config.splits_path == vak.common.converters.expanded_user_path(splits_path)
        else:
            assert dataset_config.splits_path is None
        if name is not None:
            assert dataset_config.name == name
        else:
            assert dataset_config.name is None

    @pytest.mark.parametrize(
        'config_dict',
        [
            {
                'path' :'~/datasets/BioSoundSegBench',
                'splits_path': 'splits/Bengalese-Finch-song-gy6or6-replicate-1.json',
                'name': 'BioSoundSegBench',
            },
            {
                'path' :'~/user/prepped/dataset',
            },
            {
                'path' :'~/user/prepped/dataset',
                'splits_path': 'splits/replicate-1.json'
            },
            {
                'path' :'~/user/prepped/dataset',
                'params': {'window_size': 2000}
            },
            {
                'name' : 'BioSoundSegBench',
                'path' :'~/user/prepped/dataset',
                'params': {'window_size': 2000},
            },
        ]
    )
    def test_from_config_dict(self, config_dict):
        dataset_config = vak.config.dataset.DatasetConfig.from_config_dict(config_dict)
        assert isinstance(dataset_config, vak.config.dataset.DatasetConfig)
        assert dataset_config.path == vak.common.converters.expanded_user_path(config_dict['path'])
        if 'splits_path' in config_dict:
            assert dataset_config.splits_path == vak.common.converters.expanded_user_path(config_dict['splits_path'])
        else:
            assert dataset_config.splits_path is None
        if 'name' in config_dict:
            assert dataset_config.name == config_dict['name']
        else:
            assert dataset_config.name is None
        if 'params' in config_dict:
            assert dataset_config.params == config_dict['params']
        else:
            assert dataset_config.params == {}

    @pytest.mark.parametrize(
        'config_dict',
        [
            {
                'path' :'~/datasets/BioSoundSegBench',
                'splits_path': 'splits/Bengalese-Finch-song-gy6or6-replicate-1.json',
                'name': 'BioSoundSegBench',
            },
            {
                'path' :'~/user/prepped/dataset',
            },
            {
                'path' :'~/user/prepped/dataset',
                'splits_path': 'splits/replicate-1.json'
            },
            {
                'path' :'~/user/prepped/dataset',
                'params': {'window_size': 2000}
            },
            {
                'name' : 'BioSoundSegBench',
                'path' :'~/user/prepped/dataset',
                'params': {'window_size': 2000},
            },
        ]
    )
    def test_asdict(self, config_dict):
        dataset_config = vak.config.dataset.DatasetConfig.from_config_dict(config_dict)

        dataset_config_as_dict = dataset_config.asdict()

        assert isinstance(dataset_config_as_dict, dict)
        for key in ('name', 'path', 'splits_path', 'params'):
            if key in config_dict:
                if 'path' in key:
                    assert dataset_config_as_dict[key] == vak.common.converters.expanded_user_path(config_dict[key])
                else:
                    assert dataset_config_as_dict[key] == config_dict[key]
            else:
                if key == 'params':
                    assert dataset_config_as_dict[key] == {}
                else:
                    assert dataset_config_as_dict[key] is None



"""Helper functions for setting up directories."""
import logging

from . import constants


logger = logging.getLogger(__name__)


def make_spect_output_dir_in_generated():
    constants.GENERATED_SPECT_OUTPUT_DIR.mkdir()

def make_source_files_csv_dir_in_generated():
    constants.GENERATED_SOURCE_FILES_CSV_DIR.mkdir()

def make_source_files_with_splits_csv_dir_in_generated():
    constants.GENERATED_SOURCE_FILES_WITH_SPLITS_CSV_DIR.mkdir()


def make_subdirs_in_generated(config_paths):
    """make sub-directories inside ./tests/data_for_tests/generated

    do this after copying configs,
    before using those configs to generate results.
    We use configs to decide which dirs we need to make

    makes three directories in data_for_tests/generated:
    configs, prep, and results.
    prep has one sub-directory for every data "type".
    results does also, but in addition will have sub-directories
    within those for models.
    """
    logger.info(
        "Making sub-directories in ./tests/data_for_tests/generated/ where files generated by `vak` will go"
    )

    for top_level_dir in constants.TOP_LEVEL_DIRS:  # datasets / results
        subdir_to_make = (
                constants.GENERATED_TEST_DATA_ROOT / top_level_dir
        )
        logger.info(
            f"Making sub-directory: {subdir_to_make}"
        )
        subdir_to_make.mkdir(parents=True)

    for config_metadata in constants.CONFIG_METADATA:
        config_type = config_metadata.config_type  # train, eval, predict, etc.
        if config_metadata.audio_format:
            data_dir = f"audio_{config_metadata.audio_format}_annot_{config_metadata.annot_format}"
        elif config_metadata.spect_format:
            data_dir = f"spect_{config_metadata.spect_format}_annot_{config_metadata.annot_format}"
        else:
            raise ValueError(
                f'could not determine data dir for config metadata:\n{config_metadata}'
            )
        model = config_metadata.model

        if config_metadata.use_dataset_from_config is None:  # we need to make dataset dir
            subdir_to_make = (
                    constants.GENERATED_TEST_DATA_ROOT / 'prep' / config_type / data_dir / model
            )
            logger.info(
                f"Making sub-directory: {subdir_to_make}"
            )
            subdir_to_make.mkdir(parents=True)

        subdir_to_make = (
                constants.GENERATED_TEST_DATA_ROOT / 'results' / config_type / data_dir / model
        )
        logger.info(
            f"Making sub-directory: {subdir_to_make}"
        )
        subdir_to_make.mkdir(parents=True)



import pathlib

import pytest

import vak.datasets


@pytest.mark.parametrize(
        'dataset_config, split',
        [
            (
                {
                    'name': 'BioSoundSegBench',
                    'params': {
                        'window_size': 2000,
                        'target_type': 'multi_frame_labels',
                        },
                },
                'train',
            ),
            (
                {
                    'name': 'BioSoundSegBench',
                    'params': {
                        'window_size': 2000,
                        'target_type': 'binary_frame_labels',
                        },
                },
                'train',
            ),
            (
                {
                    'name': 'BioSoundSegBench',
                    'params': {
                        'window_size': 2000,
                        'target_type': 'boundary_frame_labels',
                        },
                },
                'train',
            ),
            (
                {
                    'name': 'BioSoundSegBench',
                    'params': {
                        'window_size': 2000,
                        'target_type': 'boundary_frame_labels',
                        },
                },
                'train',
            ),
            (
                {
                    'name': 'BioSoundSegBench',
                    'params': {
                        'window_size': 2000,
                        'target_type': 'boundary_frame_labels',
                        },
                },
                'val',
            ),
            (
                {
                    'name': 'BioSoundSegBench',
                    'params': {
                        'window_size': 2000,
                        'target_type': 'boundary_frame_labels',
                        },
                },
                'test',
            ),

        ]
)
def test_get_biosoundsegbench(dataset_config, split, mock_biosoundsegbench_dataset):
    dataset_path, splits_path = mock_biosoundsegbench_dataset
    dataset_config["path"] = dataset_path
    dataset_config["splits_path"] = splits_path

    dataset = vak.datasets.get(dataset_config, split)

    assert isinstance(dataset, vak.datasets.BioSoundSegBench)
    assert dataset.dataset_path == pathlib.Path(dataset_config["path"])
    assert dataset.splits_path == pathlib.Path(dataset_config["splits_path"])
    assert dataset.window_size == dataset_config["params"]["window_size"]





"""High-level function that evaluates trained models."""

from __future__ import annotations

import logging
import pathlib

from .. import models
from ..common import validators
from .frame_classification import eval_frame_classification_model
from .parametric_umap import eval_parametric_umap_model

logger = logging.getLogger(__name__)


def eval(
    model_config: dict,
    dataset_config: dict,
    trainer_config: dict,
    checkpoint_path: str | pathlib.Path,
    output_dir: str | pathlib.Path,
    num_workers: int,
    labelmap_path: str | pathlib.Path | None = None,
    batch_size: int | None = None,
    frames_standardizer_path: str | pathlib.Path = None,
    post_tfm_kwargs: dict | None = None,
    device: str | None = None,
) -> None:
    """Evaluate a trained model.

    Parameters
    ----------
    model_config : dict
        Model configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.ModelConfig.asdict`.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    checkpoint_path : str, pathlib.Path
        path to directory with checkpoint files saved by Torch, to reload model
    output_dir : str, pathlib.Path
        Path to location where .csv files with evaluation metrics should be saved.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader. Default is 2.
    labelmap_path : str, pathlib.Path, optional
        Path to 'labelmap.json' file.
        Optional, default is None.
    batch_size : int, optional.
        Number of samples per batch fed into model.
        Optional, default is None.
    split : str
        split of dataset on which model should be evaluated.
        One of {'train', 'val', 'test'}. Default is 'test'.
    frames_standardizer_path : str, pathlib.Path
        path to a saved FramesStandardizer object used to standardize frames.
        If frames were standardized during training, and this is not provided,
        then evaluation  will give incorrect results.
        Default is None.
    post_tfm_kwargs : dict
        Keyword arguments to post-processing transform.
        If None, then no additional clean-up is applied
        when transforming labeled timebins to segments,
        the default behavior. The transform used is
        ``vak.transforms.frame_labels.PostProcess`.
        Valid keyword argument names are 'majority_vote'
        and 'min_segment_dur', and should be appropriate
        values for those arguments: Boolean for ``majority_vote``,
        a float value for ``min_segment_dur``.
        See the docstring of the transform for more details on
        these arguments and how they work.
    device : str
        Device on which to work with model + data.
        Defaults to 'cuda' if torch.cuda.is_available is True.

    Notes
    -----
    Note that unlike ``core.predict``, this function
    can modify ``labelmap`` so that metrics like edit distance
    are correctly computed, by converting any string labels
    in ``labelmap`` with multiple characters
    to (mock) single-character labels,
    with ``vak.labels.multi_char_labels_to_single_char``.
    """
    # ---- pre-conditions ----------------------------------------------------------------------------------------------
    for path, path_name in zip(
        (checkpoint_path, labelmap_path, frames_standardizer_path),
        ("checkpoint_path", "labelmap_path", "frames_standardizer_path"),
    ):
        if path is not None:  # because `frames_standardizer_path` is optional
            if not validators.is_a_file(path):
                raise FileNotFoundError(
                    f"value for ``{path_name}`` not recognized as a file: {path}"
                )

    dataset_path = pathlib.Path(dataset_config["path"])
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    model_name = model_config["name"]
    try:
        model_family = models.registry.MODEL_FAMILY_FROM_NAME[model_name]
    except KeyError as e:
        raise ValueError(
            f"No model family found for the model name specified: {model_name}"
        ) from e

    if model_family == "FrameClassificationModel":
        eval_frame_classification_model(
            model_config=model_config,
            dataset_config=dataset_config,
            trainer_config=trainer_config,
            checkpoint_path=checkpoint_path,
            labelmap_path=labelmap_path,
            output_dir=output_dir,
            num_workers=num_workers,
            frames_standardizer_path=frames_standardizer_path,
            post_tfm_kwargs=post_tfm_kwargs,
        )
    elif model_family == "ParametricUMAPModel":
        eval_parametric_umap_model(
            model_config=model_config,
            dataset_config=dataset_config,
            trainer_config=trainer_config,
            checkpoint_path=checkpoint_path,
            output_dir=output_dir,
            batch_size=batch_size,
            num_workers=num_workers,
        )
    else:
        raise ValueError(f"Model family not recognized: {model_family}")



"""Function that trains models in the Parametric UMAP family."""

from __future__ import annotations

import datetime
import logging
import pathlib

import lightning
import pandas as pd
import torch.utils.data

from .. import datapipes, models
from ..common import validators
from ..common.paths import generate_results_dir_name_as_path
from ..datapipes.parametric_umap import Datapipe

logger = logging.getLogger(__name__)


def get_split_dur(df: pd.DataFrame, split: str) -> float:
    """Get duration of a split in a dataset from a pandas DataFrame representing the dataset."""
    return df[df["split"] == split]["duration"].sum()


def get_trainer(
    accelerator: str,
    devices: int | list[int],
    max_epochs: int,
    ckpt_root: str | pathlib.Path,
    ckpt_step: int,
    log_save_dir: str | pathlib.Path,
) -> lightning.pytorch.Trainer:
    """Returns an instance of ``lightning.pytorch.Trainer``
    with a default set of callbacks.
    Used by ``vak.core`` functions."""
    ckpt_callback = lightning.pytorch.callbacks.ModelCheckpoint(
        dirpath=ckpt_root,
        filename="checkpoint",
        every_n_train_steps=ckpt_step,
        save_last=True,
        verbose=True,
    )
    ckpt_callback.CHECKPOINT_NAME_LAST = "checkpoint"
    ckpt_callback.FILE_EXTENSION = ".pt"

    val_ckpt_callback = lightning.pytorch.callbacks.ModelCheckpoint(
        monitor="val_loss",
        dirpath=ckpt_root,
        save_top_k=1,
        mode="min",
        filename="min-val-loss-checkpoint",
        auto_insert_metric_name=False,
        verbose=True,
    )
    val_ckpt_callback.FILE_EXTENSION = ".pt"

    callbacks = [
        ckpt_callback,
        val_ckpt_callback,
    ]

    logger = lightning.pytorch.loggers.TensorBoardLogger(save_dir=log_save_dir)

    trainer = lightning.pytorch.Trainer(
        max_epochs=max_epochs,
        accelerator=accelerator,
        devices=devices,
        logger=logger,
        callbacks=callbacks,
    )
    return trainer


def train_parametric_umap_model(
    model_config: dict,
    dataset_config: dict,
    trainer_config: dict,
    batch_size: int,
    num_epochs: int,
    num_workers: int,
    checkpoint_path: str | pathlib.Path | None = None,
    root_results_dir: str | pathlib.Path | None = None,
    results_path: str | pathlib.Path | None = None,
    shuffle: bool = True,
    val_step: int | None = None,
    ckpt_step: int | None = None,
    subset: str | None = None,
) -> None:
    """Train a model from the parametric UMAP family
    and save results.

    Saves checkpoint files for model,
    label map, and spectrogram scaler.
    These are saved either in ``results_path``
    if specified, or a new directory
    made inside ``root_results_dir``.

    Parameters
    ----------
    model_config : dict
        Model configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.ModelConfig.asdict`.
    dataset_config: dict
        Dataset configuration in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.DatasetConfig.asdict`.
    trainer_config: dict
        Configuration for :class:`lightning.pytorch.Trainer` in a :class:`dict`.
        Can be obtained by calling :meth:`vak.config.TrainerConfig.asdict`.
    batch_size : int
        number of samples per batch presented to models during training.
    num_epochs : int
        number of training epochs. One epoch = one iteration through the entire
        training set.
    num_workers : int
        Number of processes to use for parallel loading of data.
        Argument to torch.DataLoader.
    checkpoint_path : str, pathlib.Path, optional
        path to a checkpoint file,
        e.g., one generated by a previous run of ``vak.core.train``.
        If specified, this checkpoint will be loaded into model.
        Used when continuing training.
        Default is None, in which case a new model is initialized.
    root_results_dir : str, pathlib.Path, optional
        Root directory in which a new directory will be created
        where results will be saved.
    results_path : str, pathlib.Path, optional
        Directory where results will be saved.
        If specified, this parameter overrides ``root_results_dir``.
    val_step : int
        Computes the loss using validation set every ``val_step`` epochs.
        Default is None, in which case no validation is done.
    ckpt_step : int
        Step on which to save to checkpoint file.
        If ckpt_step is n, then a checkpoint is saved every time
        the global step / n is a whole number, i.e., when ckpt_step modulo the global step is 0.
        Default is None, in which case checkpoint is only saved at the last epoch.
    shuffle: bool
        if True, shuffle training data before each epoch. Default is True.
    split : str
        Name of split from dataset found at ``dataset_path`` to use
        when training model. Default is 'train'. This parameter is used by
        `vak.learncurve.learncurve` to specify specific subsets of the
        training set to use when training models for a learning curve.
    """
    for path, path_name in zip(
        (checkpoint_path,),
        ("checkpoint_path",),
    ):
        if path is not None:
            if not validators.is_a_file(path):
                raise FileNotFoundError(
                    f"value for ``{path_name}`` not recognized as a file: {path}"
                )

    dataset_path = pathlib.Path(dataset_config["path"])
    if not dataset_path.exists() or not dataset_path.is_dir():
        raise NotADirectoryError(
            f"`dataset_path` not found or not recognized as a directory: {dataset_path}"
        )

    logger.info(
        f"Loading dataset from path: {dataset_path}",
    )
    metadata = datapipes.parametric_umap.Metadata.from_dataset_path(
        dataset_path
    )
    dataset_csv_path = dataset_path / metadata.dataset_csv_filename
    dataset_df = pd.read_csv(dataset_csv_path)
    # ---------------- pre-conditions ----------------------------------------------------------------------------------
    if val_step and not dataset_df["split"].str.contains("val").any():
        raise ValueError(
            f"val_step set to {val_step} but dataset does not contain a validation set; "
            f"please run `vak prep` with a config.toml file that specifies a duration for the validation set."
        )

    # ---- set up directory to save output -----------------------------------------------------------------------------
    if results_path:
        results_path = pathlib.Path(results_path).expanduser().resolve()
        if not results_path.is_dir():
            raise NotADirectoryError(
                f"results_path not recognized as a directory: {results_path}"
            )
    else:
        results_path = generate_results_dir_name_as_path(root_results_dir)
        results_path.mkdir()

    # ---------------- load training data  -----------------------------------------------------------------------------
    logger.info(f"using training dataset from {dataset_path}")
    # below, if we're going to train network to predict unlabeled segments, then
    # we need to include a class for those unlabeled segments in labelmap,
    # the mapping from labelset provided by user to a set of consecutive
    # integers that the network learns to predict
    train_dur = get_split_dur(dataset_df, "train")
    logger.info(
        f"Total duration of training split from dataset (in s): {train_dur}",
    )

    dataset_params = dataset_config["params"]
    train_dataset = Datapipe.from_dataset_path(
        dataset_path=dataset_path,
        split="train",
        subset=subset,
        **dataset_params,
    )
    logger.info(
        f"Duration of ParametricUMAPDataset used for training, in seconds: {train_dataset.duration}",
    )
    train_loader = torch.utils.data.DataLoader(
        dataset=train_dataset,
        shuffle=shuffle,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    # ---------------- load validation set (if there is one) -----------------------------------------------------------
    if val_step:
        val_dataset = Datapipe.from_dataset_path(
            dataset_path=dataset_path,
            split="val",
            **dataset_params,
        )
        logger.info(
            f"Duration of ParametricUMAPDataset used for validation, in seconds: {val_dataset.duration}",
        )
        val_loader = torch.utils.data.DataLoader(
            dataset=val_dataset,
            shuffle=False,
            batch_size=batch_size,
            num_workers=num_workers,
        )
    else:
        val_loader = None

    model_name = model_config["name"]
    model = models.get(
        model_name,
        model_config,
        input_shape=train_dataset.shape,
    )

    if checkpoint_path is not None:
        logger.info(
            f"loading checkpoint for {model_name} from path: {checkpoint_path}",
        )
        model.load_state_dict_from_path(checkpoint_path)

    results_model_root = results_path.joinpath(model_name)
    results_model_root.mkdir()
    ckpt_root = results_model_root.joinpath("checkpoints")
    ckpt_root.mkdir()
    logger.info(f"training {model_name}")
    trainer = get_trainer(
        accelerator=trainer_config["accelerator"],
        devices=trainer_config["devices"],
        max_epochs=num_epochs,
        log_save_dir=results_model_root,
        ckpt_root=ckpt_root,
        ckpt_step=ckpt_step,
    )
    train_time_start = datetime.datetime.now()
    logger.info(f"Training start time: {train_time_start.isoformat()}")
    trainer.fit(
        model=model,
        train_dataloaders=train_loader,
        val_dataloaders=val_loader,
    )
    train_time_stop = datetime.datetime.now()
    logger.info(f"Training stop time: {train_time_stop.isoformat()}")
    elapsed = train_time_stop - train_time_start
    logger.info(f"Elapsed training time: {elapsed}")




import pytest

import vak.datasets


class TestBioSoundSegBench:
    @pytest.mark.parametrize(
            'split, window_size, target_type',
            [
                (
                    "train",
                    2000,
                    "multi_frame_labels",
                ),
                (
                    "train",
                    2000,
                    "binary_frame_labels",
                ),
                (
                    "train",
                    2000,
                    "boundary_frame_labels",
                ),
                (
                    "train",
                    2000,
                    ("multi_frame_labels", "boundary_frame_labels"),
                ),
            ]
    )
    def test_init(self, split, window_size, target_type, mock_biosoundsegbench_dataset):
        dataset_path, splits_path = mock_biosoundsegbench_dataset
        dataset = vak.datasets.BioSoundSegBench(
            dataset_path,
            splits_path,
            split,
            window_size,
            target_type,
        )
        assert isinstance(dataset, vak.datasets.BioSoundSegBench)

    @pytest.mark.parametrize(
            'dataset_path, splits_path, split, window_size, target_type, expected_exception',
            [
                # invalid dataset path -> NotADirectoryError
                (
                    'path/to/dataset/that/doesnt/exist',
                    None,
                    "train",
                    2000,
                    "multi_frame_labels",
                    NotADirectoryError,
                ),
                # invalid splits path -> FileNotFoundError
                (
                    None,
                    'path/to/splits/that/doesnt/exist',
                    "train",
                    2000,
                    "binary_frame_labels",
                    FileNotFoundError,
                ),
                # invalid split -> ValueError
                (
                    None,
                    None,
                    "evaluate",
                    2000,
                    "boundary_frame_labels",
                    ValueError,
                ),
                # no target type when split != "predict" -> ValueError
                (
                    None,
                    None,
                    "train",
                    2000,
                    None,
                    ValueError,
                ),
                # wrong type for target type -> TypeError
                (
                    None,
                    None,
                    "train",
                    2000,
                    1,
                    TypeError,
                ),
                # wrong type for target type -> TypeError
                (
                    None,
                    None,
                    "train",
                    2000,
                    ("boundary_frame_labels", 1),
                    TypeError,
                ),
                # invalid target type -> ValueError
                (
                    None,
                    None,
                    "train",
                    2000,
                    "frame_labels",
                    ValueError,
                ),
            ]
    )
    def test_init_raises(
            self, dataset_path, splits_path, split, window_size, target_type, expected_exception, mock_biosoundsegbench_dataset
        ):
        if dataset_path is None and splits_path is not None:
            dataset_path, _ = mock_biosoundsegbench_dataset
        elif dataset_path is not None and splits_path is None:
            _, splits_path = mock_biosoundsegbench_dataset
        elif dataset_path is None and splits_path is None:
            dataset_path, splits_path = mock_biosoundsegbench_dataset

        with pytest.raises(expected_exception):
            dataset = vak.datasets.BioSoundSegBench(
                dataset_path,
                splits_path,
                split,
                window_size,
                target_type,
            )



"""validators used by attrs-based classes and by vak.parse.parse_config"""

import pathlib

import tomlkit

from .. import models
from ..common import constants


def is_a_directory(instance, attribute, value):
    """check if given path is a directory"""
    if not pathlib.Path(value).is_dir():
        raise NotADirectoryError(
            f"Value specified for {attribute.name} of {type(instance)} not recognized as a directory:\n"
            f"{value}"
        )


def is_a_file(instance, attribute, value):
    """check if given path is a file"""
    if not pathlib.Path(value).is_file():
        raise FileNotFoundError(
            f"Value specified for {attribute.name} of {type(instance)} not recognized as a file:\n"
            f"{value}"
        )


def is_valid_model_name(instance, attribute, value: str) -> None:
    """Validate model name."""
    if value not in models.registry.MODEL_NAMES:
        raise ValueError(
            f"Invalid model name: {value}.\nValid model names are: {models.registry.MODEL_NAMES}"
        )


def is_audio_format(instance, attribute, value):
    """Check if valid audio format"""
    if value not in constants.VALID_AUDIO_FORMATS:
        raise ValueError(f"{value} is not a valid format for audio files")


def is_annot_format(instance, attribute, value):
    """Check if valid annotation format"""
    if value not in constants.VALID_ANNOT_FORMATS:
        raise ValueError(
            f"{value} is not a valid format for annotation files.\n"
            f"Valid formats are: {constants.VALID_ANNOT_FORMATS}"
        )


def is_spect_format(instance, attribute, value):
    """Check if valid format for spectrograms"""
    if value not in constants.VALID_SPECT_FORMATS:
        raise ValueError(
            f"{value} is not a valid format for spectrogram files.\n"
            f"Valid formats are: {constants.VALID_SPECT_FORMATS}"
        )


CONFIG_DIR = pathlib.Path(__file__).parent
VALID_TOML_PATH = CONFIG_DIR.joinpath("valid-version-1.2.toml")
with VALID_TOML_PATH.open("r") as fp:
    VALID_DICT = tomlkit.load(fp)["vak"]
VALID_TOP_LEVEL_TABLES = list(VALID_DICT.keys())
VALID_KEYS = {
    table_name: list(table_config_dict.keys())
    for table_name, table_config_dict in VALID_DICT.items()
}


def are_tables_valid(config_dict, toml_path=None):
    """Validate top-level tables in class:`dict`.

    This function expects the ``config_dict``
    returned by :func:`vak.config.load._load_from_toml_path`.
    """
    tables = list(config_dict.keys())
    from ..cli.cli import CLI_COMMANDS  # avoid circular import

    cli_commands_besides_prep = [
        command for command in CLI_COMMANDS if command != "prep"
    ]
    tables_that_are_commands_besides_prep = [
        table for table in tables if table in cli_commands_besides_prep
    ]
    if len(tables_that_are_commands_besides_prep) == 0:
        raise ValueError(
            "Did not find a table related to a vak command in config besides `prep`.\n"
            f"Sections in config were: {tables}\n"
            "Please see example toml configuration files here: https://github.com/vocalpy/vak/tree/main/doc/toml"
        )

    if len(tables_that_are_commands_besides_prep) > 1:
        raise ValueError(
            "Found multiple tables related to a vak command in config besides `prep`.\n"
            f"Those tables are: {tables_that_are_commands_besides_prep}. "
            f"Please use just one command besides `prep` per .toml configuration file.\n"
            "See example toml configuration files here: https://github.com/vocalpy/vak/tree/main/doc/toml"
        )

    for table in tables:
        if table not in VALID_TOP_LEVEL_TABLES:
            if toml_path:
                err_msg = (
                    f"Top-level table defined in {toml_path} is not valid: {table}\n"
                    f"Valid top-level tables are: {VALID_TOP_LEVEL_TABLES}\n"
                    "Please see example toml configuration files here: "
                    "https://github.com/vocalpy/vak/tree/main/doc/toml"
                )
            else:
                err_msg = (
                    f"Table defined in toml config is not valid: {table}\n"
                    f"Valid top-level tables are: {VALID_TOP_LEVEL_TABLES}\n"
                    "Please see example toml configuration files here: "
                    "https://github.com/vocalpy/vak/tree/main/doc/toml"
                )
            raise ValueError(err_msg)


def are_keys_valid(
    config_dict: dict,
    table_name: str,
    toml_path: str | pathlib.Path | None = None,
) -> None:
    """Given a :class:`dict` containing the *entire* configuration loaded from a toml file,
    validate the key names for a specific top-level table, e.g. ``vak.train`` or ``vak.predict``
    """
    table_keys = set(config_dict[table_name].keys())
    valid_keys = set(VALID_KEYS[table_name])
    if not table_keys.issubset(valid_keys):
        invalid_keys = table_keys - valid_keys
        if toml_path:
            err_msg = (
                f"The following keys from '{table_name}' table in "
                f"the config file '{toml_path.name}' are not valid:\n{invalid_keys}"
            )
        else:
            err_msg = (
                f"The following keys from '{table_name}' table in "
                f"the toml config are not valid:\n{invalid_keys}"
            )
        raise ValueError(err_msg)


def are_table_keys_valid(
    table_config_dict: dict,
    table_name: str,
    toml_path: str | pathlib.Path | None = None,
) -> None:
    """Given a :class:`dict` containing the configuration for a *specific* top-level table,
    loaded from a toml file, validate the key names for that table,
    e.g. ``vak.train`` or ``vak.predict``.

    This function assumes ``table_config_dict`` comes from the entire ``config_dict``
    returned by :func:`vak.config.parse.from_toml_path`, accessed using the table name as a key,
    unlike :func:`are_keys_valid`. This function is used by the ``from_config_dict``
    classmethod of the top-level tables.
    """
    table_keys = set(table_config_dict.keys())
    valid_keys = set(VALID_KEYS[table_name])
    if not table_keys.issubset(valid_keys):
        invalid_keys = table_keys - valid_keys
        if toml_path:
            err_msg = (
                f"The following keys from '{table_name}' table in "
                f"the config file '{toml_path.name}' are not valid:\n{invalid_keys}"
            )
        else:
            err_msg = (
                f"The following keys from '{table_name}' table in "
                f"the toml config are not valid:\n{invalid_keys}"
            )
        raise ValueError(err_msg)



"""Function that gets an instance of a model,
given its name and a configuration as a dict."""

from __future__ import annotations

import inspect
from typing import Callable

import lightning

from . import registry


def get(
    name: str,
    config: dict,
    input_shape: tuple[int, int, int],
    num_classes: int | None = None,
    labelmap: dict | None = None,
    post_tfm: Callable | None = None,
) -> lightning.LightningModule:
    """Get a model instance, given its name and
    a configuration as a :class:`dict`.

    Parameters
    ----------
    name : str
        Model name, must be one of vak.models.registry.MODEL_NAMES.
    config: dict
        Model configuration in a ``dict``,
        as loaded from a .toml file,
        and used by the model method ``from_config``.
    num_classes : int
        Number of classes model will be trained to classify.
    input_shape : tuple
        Of int values, sizes of dimensions,
        e.g. (channels, height, width).
        Batch size is not required for input shape.
    post_tfm : callable
        Post-processing transform that models applies during evaluation.
        Default is None, in which case the model defaults to using
        ``vak.transforms.frame_labels.ToLabels`` (that does not
        apply any post-processing clean-ups).
        To be valid, ``post_tfm`` must be either an instance of
        ``vak.transforms.frame_labels.PostProcess``.

    Returns
    -------
    model : lightning.LightningModule
        Instance of :class:`lightning.LightningModule`,
        one of the model familes.
    """
    # we do this dynamically so we always get all registered models
    try:
        model_factory = registry.MODEL_REGISTRY[name]
    except KeyError as e:
        raise ValueError(
            f"Invalid model name: '{name}'.\n"
            f"Valid model names are: {registry.MODEL_NAMES}"
        ) from e

    model_family = registry.MODEL_FAMILY_FROM_NAME[name]

    if model_family == "FrameClassificationModel":
        # still need to special case model logic here
        net_init_params = list(
            inspect.signature(
                model_factory.definition.network.__init__
            ).parameters.keys()
        )
        if ("num_input_channels" in net_init_params) and (
            "num_freqbins" in net_init_params
        ):
            num_input_channels = input_shape[-3]
            num_freqbins = input_shape[-2]
            config["network"].update(
                num_classes=num_classes,
                num_input_channels=num_input_channels,
                num_freqbins=num_freqbins,
            )
        else:
            raise ValueError(
                f"Detected that model with name '{name}' was family '{model_family}', but "
                f"unable to determine network init arguments for model. Currently all models "
                f"in this family must have networks with parameters ``num_input_channels`` and ``num_freqbins``"
            )
        model = model_factory.from_config(
            config=config, labelmap=labelmap, post_tfm=post_tfm
        )
    elif model_family == "ParametricUMAPModel":
        encoder_init_params = list(
            inspect.signature(
                model_factory.definition.network["encoder"].__init__
            ).parameters.keys()
        )
        if "input_shape" in encoder_init_params:
            if "encoder" in config["network"]:
                config["network"]["encoder"].update(input_shape=input_shape)
            else:
                config["network"]["encoder"] = dict(input_shape=input_shape)

        model = model_factory.from_config(config=config)
    else:
        raise ValueError(
            f"Value for ``model_family`` not recognized: {model_family}"
        )

    return model



#!/usr/bin/env python
# coding: utf-8
"""
Script that generates the "source" test data
used to test the implementation of the DAS model.

This script should be run in an environment
created by the 'das-test-data-env' files
in ``./tests/scripts/``.
"""
import pathlib
import urllib.request

import das.data
import das.io
import das.tcn  # need to make sure this is loaded before loading model
import das.utils
import numpy as np
from tensorflow.keras import backend as K


HERE = pathlib.Path(
    '/home/pimienta/Documents/repos/coding/bioacoustics/das-experiments/src/scripts'
)
TEST_DATA_ROOT = pathlib.Path(
    "/home/pimienta/Documents/repos/coding/vocalpy/vak-vocalpy"
) / "tests/data_for_tests"
SOURCE_TEST_DATA = TEST_DATA_ROOT / "source"
NETS_TEST_DATA = SOURCE_TEST_DATA / "nets"
DAS_TEST_DATA = NETS_TEST_DATA / "das"
STFT_KERNELS_ROOT = DAS_TEST_DATA / "stft_kernels"
DAS_BATCH_DATA_ROOT = DAS_TEST_DATA / "batches"


DAS_MENAGERIE_ROOT = DAS_TEST_DATA / 'das-menagerie'

BF_MODEL_URL = 'https://github.com/janclemenslab/das-menagerie/releases/download/bengalese_finch_v1.0/fourmales_model.h5'
BF_MODEL_PATH = DAS_MENAGERIE_ROOT / 'bengalese_finch_v1.0/fourmales_model.h5'

if not BF_MODEL_PATH.exists():
    urllib.request.urlretrieve(BF_MODEL_URL, BF_MODEL_PATH)

BF_PARAMS_URL = 'https://github.com/janclemenslab/das-menagerie/releases/download/bengalese_finch_v1.0/fourmales_params.yaml'
BF_PARAMS_PATH = DAS_MENAGERIE_ROOT / 'bengalese_finch_v1.0/fourmales_params.yaml'

if not BF_PARAMS_PATH.exists():
    urllib.request.urlretrieve(BF_PARAMS_URL, BF_PARAMS_PATH)


# TODO: I actually made this dataset, see make-datasets.ipynb -- need to be able to download!
# I think because I couldn't load directly?
data_dir='/home/pimienta/Documents/data/vocal/DAS-datasets/bengalese-finch-train-test/dataset.npy'
x_suffix: str = ""
y_suffix: str = ""
dataset = das.io.load(data_dir, x_suffix=x_suffix, y_suffix=y_suffix)


params = das.utils.load_params(
    str(BF_PARAMS_PATH.parent) + '/' + BF_PARAMS_PATH.stem.replace('_params', '')
)


model = das.utils.load_model(str(BF_MODEL_PATH.parent) + '/' + BF_MODEL_PATH.stem.replace('_model', ''),
                             model_dict={'tcn_stft': das.models.model_dict['tcn_stft']})


data_gen = das.data.AudioSequence(
    dataset["train"]["x"],
    dataset["train"]["y"],
    shuffle=False,
    shuffle_subset=None,
    first_sample=0,
    last_sample=None,
    nb_repeats=1,
    batch_processor=None,
    # TODO: which params do I need to get right output shape?
    **params,  # batch size, other parameters that apparently give us correct shape
)


N_BATCHES = 4

stem = 'bengalese-finch-train-test-dataset.npy'

for batch_ind, batch in enumerate(data_gen):
    if batch_ind > N_BATCHES - 1:
        break
    print(
        f'saving batch: {batch_ind}'
    )
    inputs, targets, masks = batch

    print(
        f'inputs shape: {inputs.shape}, targets shape: {targets.shape}'
    )
    
    for var, val in zip(('inputs', 'targets', 'masks'), (inputs, targets, masks)):
        path = DAS_BATCH_DATA_ROOT / f'{stem}-{var}-{batch_ind}'
        print(f'saving: {path}')
        np.save(path, val)
    
    out_0 = model.layers[0](inputs)

    input_conv2d = out_0[:, :, 0:1]
    input_conv2d = K.expand_dims(input_conv2d, 3)  # add a dummy dimension (channel axis)
    subsample = (model.layers[1].n_hop, 1)
    conv2d_output_real = K.conv2d(input_conv2d, model.layers[1].dft_real_kernels, 
                                  strides=subsample, padding=model.layers[1].padding, 
                                  data_format="channels_last")
    conv2d_output_imag = K.conv2d(input_conv2d, model.layers[1].dft_imag_kernels, 
                                  strides=subsample, padding=model.layers[1].padding, 
                                  data_format="channels_last")
    real_path = DAS_BATCH_DATA_ROOT / f'conv2d-output-real-{batch_ind}'
    imag_path = DAS_BATCH_DATA_ROOT / f'conv2d-output-imag-{batch_ind}'
    np.save(real_path, conv2d_output_real)
    np.save(imag_path, conv2d_output_imag)
    
    spect_mono_out = model.layers[1]._spectrogram_mono(out_0)
    spect_mono_out_path = DAS_BATCH_DATA_ROOT / f'spect-mono-out-{batch_ind}'
    np.save(spect_mono_out_path, spect_mono_out.numpy())

    # FIXME: we recapitulate call method here
    if model.layers[1].power_spectrogram != 2.0:
        amp_to_db_input = K.pow(K.sqrt(spect_mono_out), model.layers[1].power_spectrogram)
    else:
        amp_to_db_input = spect_mono_out
    amb_to_db_input_path = DAS_BATCH_DATA_ROOT / f'amp-to-db-input-{batch_ind}'
    np.save(amb_to_db_input_path, amp_to_db_input.numpy())

    if model.layers[1].return_decibel_spectrogram:
        amp_to_db_output = das.kapre.backend_keras.amplitude_to_decibel(amp_to_db_input)
        amb_to_db_output_path = DAS_BATCH_DATA_ROOT / f'amp-to-db-output-{batch_ind}'
        np.save(amb_to_db_output_path, amp_to_db_output.numpy())
    else:
        amp_to_db_output = None
    
    if amp_to_db_output is not None:
        expected_stft_out = amp_to_db_output
    else:
        expected_stft_out = amp_to_db_input
    
    stft_out = model.layers[1](out_0)
    stft_out_path = DAS_BATCH_DATA_ROOT / f'stft-layer-out-{batch_ind}'
    np.save(stft_out_path, stft_out.numpy())
    
    np.testing.assert_allclose(
        stft_out.numpy(),
        expected_stft_out.numpy()
    )



"""Fixtures having to do with source files, i.e.,
the "raw" files that go into a data set
used with neural networks
"""
import pandas as pd
import pytest

from .test_data import GENERATED_TEST_DATA_ROOT

# copied from vaktestdata.constants; could we add this to that with sys.path? or vice versa
GENERATED_SOURCE_FILES_CSV_DIR = GENERATED_TEST_DATA_ROOT / "source-files-csv"
GENERATED_SOURCE_FILES_WITH_SPLITS_CSV_DIR = GENERATED_TEST_DATA_ROOT / "source-files-with-splits-csv"


@pytest.fixture
def specific_source_files_csv_path():
    """Factory fixture that returns a specific source file csv"""
    def _specific_source_files_csv_path(
            config_type,
            model_name,
            annot_format,
            audio_format=None,
            spect_format=None,
    ):
        if audio_format and spect_format:
            raise ValueError(
                "Specify audio_format or spect_format, not both"
            )
        if audio_format:
            csv_filename = f'{model_name}_{config_type}_audio_{audio_format}_annot_{annot_format}.toml-source-files.csv'
        elif spect_format:
            csv_filename = f'{model_name}_{config_type}_spect_{spect_format}_annot_{annot_format}.toml-source-files.csv'
        csv_path = GENERATED_SOURCE_FILES_CSV_DIR / csv_filename
        return csv_path

    return _specific_source_files_csv_path


@pytest.fixture
def specific_source_files_df(
        specific_source_files_csv_path
):
    """Factory fixture that returns a specific source file csv"""
    def _specific_source_files_df(
            config_type,
            model_name,
            annot_format,
            audio_format=None,
            spect_format=None,
    ):
        csv_path = specific_source_files_csv_path(
            config_type,
            model_name,
            annot_format,
            audio_format,
            spect_format,
        )
        df = pd.read_csv(csv_path)
        return df
    return _specific_source_files_df


@pytest.fixture
def specific_source_files_with_split_csv_path():
    """Factory fixture that returns a specific source file csv"""
    def _specific_source_files_with_split_csv_path(
            config_type,
            model_name,
            annot_format,
            audio_format=None,
            spect_format=None,
    ):
        if audio_format and spect_format:
            raise ValueError(
                "Specify audio_format or spect_format, not both"
            )
        if audio_format:
            csv_filename = f'{model_name}_{config_type}_audio_{audio_format}_annot_{annot_format}.toml-source-files-with-split.csv'
        elif spect_format:
            csv_filename = f'{model_name}_{config_type}_spect_{spect_format}_annot_{annot_format}.toml-source-files-with-split.csv'
        csv_path = GENERATED_SOURCE_FILES_WITH_SPLITS_CSV_DIR / csv_filename
        return csv_path

    return _specific_source_files_with_split_csv_path


@pytest.fixture
def specific_source_files_with_split_df(
        specific_source_files_with_split_csv_path
):
    """Factory fixture that returns a specific source file csv"""
    def _specific_source_files_with_split_df(
            config_type,
            model_name,
            annot_format,
            audio_format=None,
            spect_format=None,
    ):
        csv_path = specific_source_files_with_split_csv_path(
            config_type,
            model_name,
            annot_format,
            audio_format,
            spect_format,
        )
        df = pd.read_csv(csv_path)
        return df
    return _specific_source_files_with_split_df



"""Function called by command-line interface for prep command"""

from __future__ import annotations

import pathlib
import shutil
import warnings

import tomlkit

from .. import config
from .. import prep as prep_module
from ..config.load import _load_toml_from_path
from ..config.validators import are_tables_valid


def purpose_from_toml(
    config_dict: dict, toml_path: str | pathlib.Path | None = None
) -> str:
    """Determine "purpose" from toml config,
    i.e., the command that will be run after we ``prep`` the data.

    By convention this is the other top-level table in the config file
    that correspond to a cli command besides ``[vak.prep]``, e.g. ``[vak.train]``.
    """
    # validate, make sure there aren't multiple commands in one config file first
    are_tables_valid(config_dict, toml_path=toml_path)
    config_dict = config_dict

    from ..cli.cli import CLI_COMMANDS  # avoid circular imports

    commands_that_are_not_prep = [
        command for command in CLI_COMMANDS if command != "prep"
    ]
    purpose = None
    for table_name in commands_that_are_not_prep:
        if table_name in config_dict:
            purpose = (
                table_name  # this top-level table is the "purpose" of the file
            )
    if purpose is None:
        raise ValueError(
            "Did not find a top-level table in configuration file that corresponds to a CLI command. "
            f"Configuration file path: {toml_path}\n"
            f"Found the following top-level tables: {config_dict.keys()}\n"
            f"Valid CLI commands besides ``prep`` (that correspond top-level tables) are: {commands_that_are_not_prep}"
        )
    return purpose


# note NO LOGGING -- we configure logger inside `core.prep`
# so we can save log file inside dataset directory

# see https://github.com/NickleDave/vak/issues/334
TABLES_PREP_SHOULD_PARSE = "prep"


def prep(toml_path):
    """Prepare datasets from vocalizations.
    Function called by command-line interface.

    Parameters
    ----------
    toml_path : str, pathlib.Path
        path to a configuration file in TOML format.
        Used to rewrite file with options determined by this function and needed for other functions

    Notes
    -----
    Saves a .csv file representing the dataset generated from data_dir.

    Datasets are used to train neural networks that segment audio files into
    vocalizations, and then predict labels for those segments.
    The function also prepares datasets so neural networks can predict the
    segmentation and annotation of vocalizations in them.
    It can also split a dataset into training, validation, and test sets,
    e.g. for benchmarking different neural network architectures.

    If the 'purpose' is set to 'train' or 'learncurve', and/or
    the duration of either the training or test set is provided,
    then the function attempts to split the dataset into training and test sets.
    A duration can also be specified for a validation set
    (used to measure performance during training).
    In these cases, the 'split' column in the .csv
    identifies which files (rows) belong to the training, test, and
    validation sets created from that Dataset.

    If the 'purpose' is set to 'predict' or 'eval',
    or no durations for any of the training sets are specified,
    then the function assumes all the vocalizations constitute a single
    dataset, and for all rows the 'split' columns for that dataset
    will be 'predict' or 'test' (respectively).
    """
    toml_path = pathlib.Path(toml_path)

    # open here because need to check whether the `dataset` already has a `path`, see #314 & #333
    config_dict = _load_toml_from_path(toml_path)

    # ---- figure out purpose of config file from tables; will save path of prep'd dataset in that table ---------------
    purpose = purpose_from_toml(config_dict, toml_path)
    if (
        "dataset" in config_dict[purpose]
        and "path" in config_dict[purpose]["dataset"]
    ):
        raise ValueError(
            f"This configuration file already has a '{purpose}.dataset' table with a 'path' key, "
            f"and running `prep` would overwrite the value for that key. To `prep` a new dataset, please "
            "either create a new configuration file, or remove "
            f"the 'path' key-value pair from the '{purpose}.dataset' table in the file:\n{toml_path}"
        )

    # now that we've checked that, go ahead and parse just the prep tabel;
    # we don't load the 'purpose' table into a config, to avoid error messages like non-existent paths, etc.
    # see https://github.com/NickleDave/vak/issues/334 and https://github.com/NickleDave/vak/issues/314
    cfg = config.Config.from_toml_path(
        toml_path, tables_to_parse=TABLES_PREP_SHOULD_PARSE
    )
    if cfg.prep is None:
        raise ValueError(
            f"prep called with a config.toml file that does not have a [vak.prep] table: {toml_path}"
        )

    if purpose == "predict":
        if cfg.prep.labelset is not None:
            warnings.warn(
                "config has a [vak.predict] table, but labelset option is specified in [vak.prep] table."
                "This would cause an error because the dataframe.from_files method will attempt to "
                f"check whether the files in the data_dir ({cfg.prep.data_dir}) have labels in "
                "labelset, even though those files don't have annotation.\n"
                "Setting labelset to None."
            )
            cfg.prep.labelset = None

    _, dataset_path = prep_module.prep(
        data_dir=cfg.prep.data_dir,
        purpose=purpose,
        dataset_type=cfg.prep.dataset_type,
        input_type=cfg.prep.input_type,
        audio_format=cfg.prep.audio_format,
        spect_format=cfg.prep.spect_format,
        spect_params=cfg.prep.spect_params,
        annot_format=cfg.prep.annot_format,
        annot_file=cfg.prep.annot_file,
        labelset=cfg.prep.labelset,
        audio_dask_bag_kwargs=cfg.prep.audio_dask_bag_kwargs,
        output_dir=cfg.prep.output_dir,
        train_dur=cfg.prep.train_dur,
        val_dur=cfg.prep.val_dur,
        test_dur=cfg.prep.test_dur,
        train_set_durs=cfg.prep.train_set_durs,
        num_replicates=cfg.prep.num_replicates,
    )

    # we re-open config using tomlkit so we can add path to dataset table in style-preserving way
    with toml_path.open("r") as fp:
        tomldoc = tomlkit.load(fp)
    if "dataset" not in tomldoc["vak"][purpose]:
        dataset_table = tomlkit.table()
        tomldoc["vak"][purpose].add("dataset", dataset_table)
    tomldoc["vak"][purpose]["dataset"].add("path", str(dataset_path))
    with toml_path.open("w") as fp:
        tomlkit.dump(tomldoc, fp)

    # lastly, copy config to dataset directory root
    shutil.copy(src=toml_path, dst=dataset_path)



from . import constants, helper
from .infer_datapipe import InferDatapipe
from .metadata import Metadata
from .train_datapipe import TrainDatapipe

__all__ = ["constants", "helper", "Metadata", "InferDatapipe", "TrainDatapipe"]



"""Class forms of transformations
related to frame labels,
i.e., vectors where each element represents
a label for a frame, either a single sample in audio
or a single time bin from a spectrogram.

These classes call functions from
``vak.transforms.frame_labels.functional``.
Not all functions in that module
have a corresponding class,
just key functions needed by
dataloaders and models.

- FromSegments: transform to get frame labels from annotations
- ToLabels: transform to get back just string labels from frame labels,
  used to evaluate a model.
- ToSegments: transform to get segment onsets, offsets, and labels from frame labels.
    Used to convert model output to predictions.
    Inverse of ``from_segments``.
- PostProcess: combines two post-processing transforms applied to frame labels,
  ``remove_short_segments`` and ``take_majority_vote``, in one class.
"""

from __future__ import annotations

import numpy as np
import numpy.typing as npt

from . import functional as F


class FromSegments:
    """Transform that makes a vector of frame labels,
    given labeled segments in the form of onset times,
    offset times, and segment labels.

    Attributes
    ----------
    background_label : int
        Label assigned to time bins that do not have labels associated with them.
        Default is 0.
    """

    def __init__(self, background_label: int = 0):
        self.background_label = background_label

    def __call__(
        self,
        labels_int: np.ndarray,
        onsets_s: np.ndarray,
        offsets_s: np.ndarray,
        time_bins: np.ndarray,
    ) -> np.ndarray:
        """Make a vector of frame labels,
        given labeled segments in the form of onset times,
        offset times, and segment labels.

        Parameters
        ----------
        labels_int : list, numpy.ndarray
            A list or array of labels from the annotation for a vocalization,
            mapped to integers
        onsets_s : numpy.ndarray
            1-d vector of floats, segment onsets in seconds.
        offsets_s : numpy.ndarray
            1-d vector of floats, segment offsets in seconds.
        time_bins : numpy.ndarray
            1-d vector of floats, time in seconds for center of each time bin of a spectrogram.

        Returns
        -------
        frame_labels : numpy.ndarray
            same length as time_bins, with each element a label for each time bin
        """
        return F.from_segments(
            labels_int,
            onsets_s,
            offsets_s,
            time_bins,
            background_label=self.background_label,
        )


class ToLabels:
    """Transforms that converts
    vector of frame labels to a string,
    one character for each continuous segment.

    Allows for converting output of network
    from a label for each frame
    to one label for each continuous segment,
    in order to compute string-based metrics like edit distance.

    Attributes
    ----------
    labelmap : dict
        That maps string labels to integers.
        The mapping is inverted to convert back to string labels.
    """

    def __init__(self, labelmap: dict):
        self.labelmap = labelmap

    def __call__(self, frame_labels: np.ndarray) -> str:
        """Convert vector of frame labels to a string,
        one character for each continuous segment.

        Parameters
        ----------
        frame_labels : numpy.ndarray
            A vector where each element represents
            a label for a frame, either a single sample in audio
            or a single time bin from a spectrogram.
            Typically, the output of a neural network.

        Returns
        -------
        labels : str
            The label at the onset of each continuous segment
            in ``frame_labels``, mapped back to string labels in ``labelmap``.
        """
        return F.to_labels(frame_labels, self.labelmap)


class ToSegments:
    """Transform that converts a vector of frame labels
    into segments in the form of onset indices,
    offset indices, and labels.

    Finds where continuous runs of a single label start
    and stop in timebins, and considers each of these runs
    a segment.

    The function returns vectors of labels and onsets and offsets
    in units of seconds.

    Attributes
    ----------
    labelmap : dict
        That maps string labels to integers.
        The mapping is inverted to convert back to string labels.
    n_decimals_trunc : int
        Number of decimal places to keep when truncating the timebin duration
        calculated from the vector of times t. Default is 5.
    """

    def __init__(self, labelmap: dict, n_decimals_trunc: int = 5):
        self.labelmap = labelmap
        self.n_decimals_trunc = n_decimals_trunc

    def __call__(
        self, frame_labels: np.ndarray, frame_times: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Convert a vector of frame labels
        into segments in the form of onset indices,
        offset indices, and labels.

        Finds where continuous runs of a single label start
        and stop in timebins, and considers each of these runs
        a segment.

        The function returns vectors of labels and onsets and offsets
        in units of seconds.

        Parameters
        ----------
        frame_labels : numpy.ndarray
            A vector where each element represents
            a label for a frame, either a single sample in audio
            or a single time bin from a spectrogram.
            Output of a neural network.
        frame_times : numpy.ndarray
            Vector of times; the times are either the time of samples in audio,
            or the bin centers of columns in a spectrogram,
            returned by function that generated spectrogram.
            Used to convert onset and offset indices in frame_labels to seconds.

        Returns
        -------
        labels : numpy.ndarray
            Vector where each element is a label for a segment with its onset
            and offset indices given by the corresponding element in onset_inds
            and offset_inds.
        onsets_s : numpy.ndarray
            Vector where each element is the onset in seconds a segment.
            Each onset corresponds to the value at the same index in labels.
        offsets_s : numpy.ndarray
            Vector where each element is the offset in seconds of a segment.
            Each offset corresponds to the value at the same index in labels.
        """
        return F.to_segments(
            frame_labels, self.labelmap, frame_times, self.n_decimals_trunc
        )


class PostProcess:
    """Apply post-processing transformations
    to a vector of frame labels.

    Optional post-processing
    consist of two transforms,
    that both rely on there being a label
    that corresponds to the background class.
    The first removes any segments that are
    shorter than a specified duration,
    by converting labels in those segments to the
    background class label.
    The second performs a "majority vote"
    transform within run of labels that is
    bordered on both sides by the "background" label.
    I.e., it counts the number of times any
    label occurs in that segment,
    and then assigns all bins the most common label.

    The function performs those steps in this order
    (pseudo-code):

    .. code-block::

       if min_segment_dur:
           frame_labels = remove_short_segments(frame_labels, labelmap, min_segment_dur)
       if majority_vote:
           frame_labels = majority_vote(frame_labels, labelmap)
       return frame_labels

    Attributes
    ----------
    timebin_dur : float
        Duration of a time bin in a spectrogram,
        e.g., as estimated from vector of times
        using ``vak.timebins.timebin_dur_from_vec``.
    background_label : int
        Label that was given to segments that were not labeled in annotation,
        e.g. silent periods between annotated segments. Default is 0.
    min_segment_dur : float
        Minimum duration of segment, in seconds. If specified, then
        any segment with a duration less than min_segment_dur is
        removed from frame_labels. Default is None, in which case no
        segments are removed.
    majority_vote : bool
        If True, transform segments containing multiple labels
        into segments with a single label by taking a "majority vote",
        i.e. assign all time bins in the segment the most frequently
        occurring label in the segment. This transform can only be
        applied if the labelmap contains an 'unlabeled' label,
        because unlabeled segments makes it possible to identify
        the labeled segments. Default is False.
    """

    def __init__(
        self,
        timebin_dur: float,
        background_label: int = 0,
        min_segment_dur: float | None = None,
        majority_vote: bool = False,
    ):
        self.timebin_dur = timebin_dur
        self.background_label = background_label
        self.min_segment_dur = min_segment_dur
        self.majority_vote = majority_vote

    def __call__(self, frame_labels: np.ndarray, boundary_labels: npt.NDArray | None = None) -> np.ndarray:
        """Apply post-processing transformations
        to a vector of frame labels.

        Parameters
        ----------
        frame_labels : numpy.ndarray
            A vector where each element represents
            a label for a frame, either a single sample in audio
            or a single time bin from a spectrogram.
            Output of a neural network.

        Returns
        -------
        frame_labels : numpy.ndarray
            Vector of frame labels after post-processing is applied.
        """
        return F.postprocess(
            frame_labels,
            self.timebin_dur,
            self.background_label,
            self.min_segment_dur,
            self.majority_vote,
            boundary_labels,
        )



from . import functional, transforms
from .functional import *  # noqa : F401
from .transforms import ToLabels  # noqa : F401
from .transforms import FromSegments, PostProcess, ToSegments

__all__ = [
    "functional",
    "transforms",
    "FromSegments",
    "PostProcess",
    "ToLabels",
    "ToSegments",
]



import attrs


@attrs.define
class ConfigMetadata:
    """Dataclass that represents metadata
    about a configuration file

    Attributes
    ----------
    filename : str
        The name of the configuration file.
    model : str
        The name of the model in :mod:`vak`
        that the configuration file is used with.
    model_family : str
        The name of the model family
        for the model in the configuration file.
    config_type : str
        The type of config, one of
        {'train', 'eval', 'predict', 'learncurve'}.
    audio_format : str
        The format of the audio files.
    spect_format : str
        The format of the spectrogram files.
    spect_output_dir : str, optional
        The directory where spectrograms should be saved
        when generated for this configuration file.
        If not specified, then no spectrograms are generated.
        This attribute is used to avoid repeatedly
        generating the same set of spectrograms for multiple
        configs.
    data_dir : str, optional
        The directory that should be used as the `data_dir`
        option for this config.
        The option will be changed to this value in the generated
        config file.
        This attribute is used to avoid repeatedly
        generating the same set of spectrograms for multiple
        configs.
    use_dataset_from_config : str, optional
        The filename of another configuration file.
        The ``dataset_path`` option of that configuration file
        will be used for this configuration file.
        This option is used to avoid repeatedly
        generating the same dataset for multiple configs.
    use_results_from_config : str, optional
        The filename of another configuration file.
        The most recent results from ``results_path`` option
        of that configuration file
        will be used for this configuration file.
    """
    filename: str = attrs.field(converter=str)
    model: str = attrs.field(converter=str)
    model_family: str = attrs.field(converter=str)
    config_type: str = attrs.field(converter=str)
    audio_format: str = attrs.field(converter=attrs.converters.optional(str), default=None)
    spect_format: str = attrs.field(converter=attrs.converters.optional(str), default=None)
    annot_format: str = attrs.field(converter=attrs.converters.optional(str), default=None)
    spect_output_dir: str = attrs.field(converter=attrs.converters.optional(str), default=None)
    data_dir: str = attrs.field(converter=attrs.converters.optional(str), default=None)
    use_dataset_from_config: str = attrs.field(converter=attrs.converters.optional(str), default=None)
    use_result_from_config: str = attrs.field(converter=attrs.converters.optional(str), default=None)



import torch

from ..nn.modules import Conv2dTF, NormReLU


class ED_TCN(torch.nn.Module):
    """Encoder-Decoder Temporal Convolutional Network.
    As described in [1]_.

    Note that this network adds convolutional layers on the front end
    to provide features fed into the ED-TCN described in [1]_.

    This implementation in PyTorch is adapted from the original in Keras
    under MIT license.
    https://github.com/colincsl/TemporalConvolutionalNetworks/blob/cccdf868ed4c7a56745b41d30fd9f1dc637eb3f3/code/tf_models.py#L68
    https://github.com/colincsl/TemporalConvolutionalNetworks/blob/master/LICENSE

    .. [1] Lea, C., Flynn, M. D., Vidal, R., Reiter, A., & Hager, G. D. (2017).
       Temporal convolutional networks for action segmentation and detection.
       In proceedings of the IEEE Conference on Computer Vision and Pattern Recognition (pp. 156-165).
    """

    def __init__(
        self,
        num_classes,
        num_input_channels=1,
        num_freqbins=256,
        padding="SAME",
        conv2d_1_filters=32,
        conv2d_1_kernel_size=(5, 5),
        conv2d_2_filters=64,
        conv2d_2_kernel_size=(5, 5),
        pool1_size=(8, 1),
        pool1_stride=(8, 1),
        pool2_size=(8, 1),
        pool2_stride=(8, 1),
        conv1d_1_filters=64,
        conv1d_2_filters=96,
        conv1d_kernel_size=25,
    ):
        super().__init__()
        self.num_classes = num_classes
        self.num_input_channels = num_input_channels
        self.num_freqbins = num_freqbins

        self.cnn = torch.nn.Sequential(
            Conv2dTF(
                in_channels=self.num_input_channels,
                out_channels=conv2d_1_filters,
                kernel_size=conv2d_1_kernel_size,
                padding=padding,
            ),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(kernel_size=pool1_size, stride=pool1_stride),
            Conv2dTF(
                in_channels=conv2d_1_filters,
                out_channels=conv2d_2_filters,
                kernel_size=conv2d_2_kernel_size,
                padding=padding,
            ),
            torch.nn.ReLU(inplace=True),
            torch.nn.MaxPool2d(kernel_size=pool2_size, stride=pool2_stride),
        )

        # determine number of features in output after stacking channels
        # we use the same number of features for hidden states
        # note self.num_hidden is also used to reshape output of cnn in self.forward method
        # determine number of features in output after stacking channels
        N_DUMMY_TIMEBINS = (
            256  # some not-small number. This dimension doesn't matter here
        )
        batch_shape = (
            1,
            self.num_input_channels,
            self.num_freqbins,
            N_DUMMY_TIMEBINS,
        )
        tmp_tensor = torch.rand(batch_shape)
        tmp_out = self.cnn(tmp_tensor)
        channels_out, freqbins_out = tmp_out.shape[1], tmp_out.shape[2]
        self.n_cnn_features_out = channels_out * freqbins_out

        self.encoder = torch.nn.Sequential(
            torch.nn.Conv1d(
                self.n_cnn_features_out,
                conv1d_1_filters,
                conv1d_kernel_size,
                padding="same",
            ),
            torch.nn.Dropout1d(p=0.3),
            NormReLU(),
            torch.nn.MaxPool1d(kernel_size=2),
            torch.nn.Conv1d(
                conv1d_1_filters,
                conv1d_2_filters,
                conv1d_kernel_size,
                padding="same",
            ),
            torch.nn.Dropout1d(0.3),
            NormReLU(),
            torch.nn.MaxPool1d(kernel_size=2),
        )

        self.decoder = torch.nn.Sequential(
            torch.nn.Upsample(scale_factor=2),
            torch.nn.Conv1d(
                conv1d_2_filters,
                conv1d_2_filters,
                conv1d_kernel_size,
                padding="same",
            ),
            torch.nn.Dropout1d(p=0.3),
            NormReLU(),
            torch.nn.Upsample(scale_factor=2),
            torch.nn.Conv1d(
                conv1d_2_filters,
                conv1d_1_filters,
                conv1d_kernel_size,
                padding="same",
            ),
            torch.nn.Dropout1d(0.3),
            NormReLU(),
        )

        self.fc = torch.nn.Linear(
            in_features=conv1d_1_filters, out_features=self.num_classes
        )

    def forward(self, x):
        x = self.cnn(x)
        # stack channels, to give tensor shape (batch, cnn features, time bins)
        x = x.view(x.shape[0], self.n_cnn_features_out, -1)
        x = self.encoder(x)
        x = self.decoder(x)
        x = x.permute(
            0, 2, 1
        )  # so that we can project features down on to number of classes
        x = self.fc(x)
        x = x.permute(
            0, 2, 1
        )  # switch back to (batch, classes, time) for loss function
        return x



from __future__ import annotations

import logging
import pathlib

import attrs
import crowsetta
import pandas as pd

from ...common import annotation, constants
from ...common.converters import expanded_user_path, labelset_to_set
from ...config.spect_params import SpectParamsConfig
from . import audio_helper, spect_helper

logger = logging.getLogger(__name__)


def prep_spectrogram_dataset(
    data_dir: str | pathlib.Path,
    annot_format: str | None = None,
    labelset: set | None = None,
    annot_file: str | pathlib.Path | None = None,
    audio_format: str | None = None,
    spect_format: str | None = None,
    spect_params: dict | None = None,
    spect_output_dir: str | pathlib.Path | None = None,
    audio_dask_bag_kwargs: dict | None = None,
) -> pd.DataFrame:
    """Make a dataset of spectrograms,
    optionally paired with annotations.

    Prepares dataset of vocalizations from a directory of audio or spectrogram files,
    and (optionally) annotation for those files. The dataset is returned as a pandas DataFrame.

    Datasets are used to train neural networks, predicting annotations for
    the dataset itself using a trained neural network, etc.

    If dataset is created from audio files, then array files containing spectrograms
    will be generated from the audio files and saved in ``spect_output_dir``
    with the extension ``.spect.npz``. The ``spect_output_dir`` defaults to ``data_dir``
    if is not specified.

    Parameters
    ----------
    data_dir : str
        path to directory with audio or spectrogram files from which to make dataset
    annot_format : str
        format of annotations. Any format that can be used with the
        crowsetta library is valid. Default is None.
    labelset : str, list, set
        of str or int, set of unique labels for vocalizations. Default is None.
        If not None, then files will be skipped where the associated annotation
        contains labels not found in ``labelset``.
        ``labelset`` is converted to a Python ``set`` using ``vak.converters.labelset_to_set``.
        See help for that function for details on how to specify labelset.
    load_spects : bool
        if True, load spectrograms. If False, return a InferDatapipe without spectograms loaded.
        Default is True. Set to False when you want to create a InferDatapipe for use
        later, but don't want to load all the spectrograms into memory yet.
    audio_format : str
        format of audio files. One of {'wav', 'cbin'}.
    spect_format : str
        format of array files containing spectrograms as 2-d matrices.
        One of {'mat', 'npz'}.
    annot_file : str
        Path to a single annotation file. Default is None.
        Used when a single file contains annotations for multiple audio files.
    spect_params : dict, vak.config.spect.SpectParamsConfig.
        Parameters for creating spectrograms.
        Default is None (implying that spectrograms are already made).
    spect_output_dir : str
        Path to location where spectrogram files should be saved.
        Default is None, in which case it defaults to ``data_dir``.
    audio_dask_bag_kwargs : dict
        Keyword arguments used when calling ``dask.bag.from_sequence``
        inside ``vak.io.audio``, where it is used to parallelize
        the conversion of audio files into spectrograms.
        Option should be specified in config.toml file as an inline table,
        e.g., ``audio_dask_bag_kwargs = { npartitions = 20 }``.
        Allows for finer-grained control
        when needed to process files of different sizes.

    Returns
    -------
    source_files_df : pandas.DataFrame
        A set of source files that will be used to prepare a
        data set for use with neural network models,
        represented as a :class:`pandas.DataFrame`.
        Will contain paths to spectrogram files,
        possibly paired with annotation files,
        as well as the original audio files if the
        spectrograms were generated from audio by
        :func:`vak.prep.audio_helper.make_spectrogram_files_from_audio_files`.
        The columns of the dataframe are specified by
        :const:`vak.prep.spectrogram_dataset.spect_helper.DF_COLUMNS`.
    """
    # ---- pre-conditions ----------------------------------------------------------------------------------------------
    if labelset is not None:
        labelset = labelset_to_set(labelset)

    if audio_format is None and spect_format is None:
        raise ValueError("Must specify either audio_format or spect_format")

    if audio_format and spect_format:
        raise ValueError(
            "Cannot specify both audio_format and spect_format, "
            "unclear whether to create spectrograms from audio files or "
            "use already-generated spectrograms from array files"
        )

    data_dir = expanded_user_path(data_dir)
    if not data_dir.is_dir():
        raise NotADirectoryError(f"data_dir not found: {data_dir}")

    if spect_output_dir:
        spect_output_dir = expanded_user_path(spect_output_dir)
        if not spect_output_dir.is_dir():
            raise NotADirectoryError(
                f"spect_output_dir not found: {spect_output_dir}"
            )
    else:
        spect_output_dir = data_dir

    if annot_format is not None:
        if annot_file is None:
            annot_files = annotation.files_from_dir(
                annot_dir=data_dir, annot_format=annot_format
            )
            scribe = crowsetta.Transcriber(format=annot_format)
            annot_list = [
                scribe.from_file(annot_file).to_annot()
                for annot_file in annot_files
            ]
        else:
            scribe = crowsetta.Transcriber(format=annot_format)
            annot_list = scribe.from_file(annot_file).to_annot()
        if isinstance(annot_list, crowsetta.Annotation):
            # if e.g. only one annotated audio file in directory, wrap in a list to make iterable
            # fixes https://github.com/NickleDave/vak/issues/467
            annot_list = [annot_list]
    else:  # if annot_format not specified
        annot_list = None

    # ------ if making dataset from audio files, need to make into array files first! ----------------------------------
    if audio_format:
        logger.info(
            f"making array files containing spectrograms from audio files in: {data_dir}",
        )
        audio_files = audio_helper.files_from_dir(data_dir, audio_format)

        spect_files = audio_helper.make_spectrogram_files_from_audio_files(
            audio_format=audio_format,
            spect_params=spect_params,
            output_dir=spect_output_dir,
            audio_files=audio_files,
            annot_list=annot_list,
            annot_format=annot_format,
            labelset=labelset,
            dask_bag_kwargs=audio_dask_bag_kwargs,
        )
        spect_format = "npz"
        spect_ext = constants.SPECT_NPZ_EXTENSION
    else:  # if audio format is None
        spect_files = None
        # make sure we use the vak extension for spectrogram files
        spect_ext = constants.SPECT_FORMAT_EXT_MAP[spect_format]

    make_dataframe_kwargs = {
        "spect_format": spect_format,
        "labelset": labelset,
        "annot_list": annot_list,
        "annot_format": annot_format,
        "spect_ext": spect_ext,
    }

    if (
        spect_files
    ):  # because we just made them, and put them in spect_output_dir
        make_dataframe_kwargs["spect_files"] = spect_files
        logger.info(
            f"creating dataset from spectrogram files in: {spect_output_dir}",
        )
    else:
        make_dataframe_kwargs["spect_dir"] = data_dir
        logger.info(
            f"creating dataset from spectrogram files in: {data_dir}",
        )

    if spect_params:  # get relevant keys for accessing arrays from array files
        if isinstance(spect_params, SpectParamsConfig):
            spect_params = attrs.asdict(spect_params)
        for key in [
            "freqbins_key",
            "timebins_key",
            "spect_key",
            "audio_path_key",
        ]:
            make_dataframe_kwargs[key] = spect_params[key]

    source_files_df = spect_helper.make_dataframe_of_spect_files(
        **make_dataframe_kwargs
    )
    return source_files_df



import inspect

import torch
import pytest

import vak.nets


class TestED_TCN:

    @pytest.mark.parametrize(
        'num_classes, num_input_channels, num_freqbins',
        [
            (
                    1, 10, None,
            ),
            (
                    1, 23, None,
            ),
            (
                    6, 1, 513
            ),
            (
                    23, 1, 512
            ),
        ]
    )
    def test_init(self, num_classes, num_input_channels, num_freqbins):
        """test we can instantiate ED_TCN
        and it has the expected attributes"""
        if num_input_channels is None or num_freqbins is None:
            init_sig = inspect.signature(vak.nets.ED_TCN.__init__)
            if num_input_channels is None:
                num_input_channels = init_sig.parameters['num_input_channels'].default
            if num_freqbins is None:
                num_freqbins = init_sig.parameters['num_freqbins'].default

        net = vak.nets.ED_TCN(num_classes, num_input_channels, num_freqbins)
        assert isinstance(net, vak.nets.ED_TCN)
        for expected_attr, expected_type in (
            ('num_classes', int),
            ('num_input_channels', int),
            ('num_freqbins', int),
            ('cnn', torch.nn.Module),
            ('encoder', torch.nn.Module),
            ('decoder', torch.nn.Module),
            ('fc', torch.nn.Linear)
        ):
            assert hasattr(net, expected_attr)
            assert isinstance(getattr(net, expected_attr), expected_type)

        assert net.num_classes == num_classes
        assert net.num_input_channels == num_input_channels
        assert net.num_freqbins == num_freqbins

    @pytest.mark.parametrize(
        'num_classes, num_input_channels, num_freqbins, num_timebins, batch_size',
        [
            (
                    10, None, None, 100, 8
            ),
            (
                    23, None, None, 100, 64
            ),
            (
                    23, 1, 512, 100, 64
            ),
        ]
    )
    def test_forward(self, num_classes, num_input_channels, num_freqbins, num_timebins, batch_size):
        """test we can forward a tensor through a ED_TCN instance
        and get the expected output"""
        if num_input_channels is None or num_freqbins is None:
            init_sig = inspect.signature(vak.nets.ED_TCN.__init__)
            if num_input_channels is None:
                num_input_channels = init_sig.parameters['num_input_channels'].default
            if num_freqbins is None:
                num_freqbins = init_sig.parameters['num_freqbins'].default

        input = torch.rand(batch_size, num_input_channels, num_freqbins, num_timebins)  # a "batch"
        net = vak.nets.ED_TCN(num_classes, num_input_channels, num_freqbins)
        out = net(input)
        assert isinstance(out, torch.Tensor)
        assert out.shape == (batch_size, num_classes, num_timebins)



import pytest

import vak

from .test_tweetynet import TEST_INIT_ARGVALS


class TestED_TCN:
    @pytest.mark.parametrize(
        'labelmap, input_shape',
        TEST_INIT_ARGVALS
    )
    def test_init(self, labelmap, input_shape):
        # network has required args that need to be determined dynamically
        num_input_channels = input_shape[-3]
        num_freqbins = input_shape[-2]
        network = vak.models.ED_TCN.definition.network(len(labelmap), num_input_channels, num_freqbins)
        model = vak.models.ED_TCN.from_instances(network=network, labelmap=labelmap)
        assert isinstance(model, vak.models.FrameClassificationModel)
        for attr in ('network', 'loss', 'optimizer'):
            assert hasattr(model, attr)
            assert isinstance(getattr(model, attr),
                              getattr(vak.models.ed_tcn.ED_TCN.definition, attr))
        assert hasattr(model, 'metrics')
        assert isinstance(model.metrics, dict)
        for metric_name, metric_callable in model.metrics.items():
            assert isinstance(metric_callable,
                              vak.models.ed_tcn.ED_TCN.definition.metrics[metric_name])



"""tests for vak.cli.eval module"""
from unittest import mock

import pytest

import vak.cli.eval
import vak.config
import vak.common.constants
import vak.common.paths

from . import cli_asserts


@pytest.mark.parametrize(
    "model_name, audio_format, spect_format, annot_format",
    [
        ("TweetyNet", "cbin", None, "notmat"),
    ],
)
def test_eval(
    model_name, audio_format, spect_format, annot_format, specific_config_toml_path, tmp_path, trainer_table
):
    output_dir = tmp_path.joinpath(
        f"test_eval_{audio_format}_{spect_format}_{annot_format}"
    )
    output_dir.mkdir()

    keys_to_change = [
        {"table": "eval", "key": "output_dir", "value": str(output_dir)},
        {"table": "eval", "key": "trainer", "value": trainer_table},
    ]

    toml_path = specific_config_toml_path(
        config_type="eval",
        model=model_name,
        audio_format=audio_format,
        annot_format=annot_format,
        spect_format=spect_format,
        keys_to_change=keys_to_change,
    )

    with mock.patch('vak.eval.eval', autospec=True) as mock_core_eval:
        vak.cli.eval.eval(toml_path)
        assert mock_core_eval.called

    assert cli_asserts.log_file_created(command="eval", output_path=output_dir)

    assert cli_asserts.log_file_contains_version(command="eval", output_path=output_dir)


def test_eval_dataset_none_raises(
        specific_config_toml_path
):
    """Test that cli.eval raises ValueError when dataset is None
    (presumably because `vak prep` was not run yet)
    """
    keys_to_change = [
        {"table": "eval", "key": "dataset", "value": "DELETE-KEY"},
    ]

    toml_path = specific_config_toml_path(
        config_type="eval",
        model="TweetyNet",
        audio_format="cbin",
        annot_format="notmat",
        spect_format=None,
        keys_to_change=keys_to_change,
    )

    with pytest.raises(KeyError):
        vak.cli.eval.eval(toml_path)



"""tests for vak.prep.spectrogram_dataset module"""
import pathlib

import pandas as pd
import pytest

import vak.common.annotation
import vak.common.constants
import vak.prep.spectrogram_dataset.prep
import vak.prep.spectrogram_dataset.spect_helper

from ...fixtures.annot import ANNOT_FILE_YARDEN
from ...fixtures.audio import AUDIO_DIR_CBIN
from ...fixtures.spect import SPECT_DIR_MAT


def assert_returned_dataframe_matches_expected(
    df_returned,
    data_dir,
    labelset,
    annot_format,
    audio_format,
    spect_format,
    spect_output_dir=None,
    annot_file=None,
    expected_audio_paths=None,
    not_expected_audio_paths=None,
    spect_file_ext=".spect.npz",
    expected_spect_paths=None,
    not_expected_spect_paths=None,
):
    """tests that dataframe returned by ``vak.prep.spectrogram_dataset.prep.prep_spectrogram_dataset``
    matches expected dataframe."""
    assert isinstance(df_returned, pd.DataFrame)

    assert df_returned.columns.values.tolist() == vak.prep.spectrogram_dataset.spect_helper.DF_COLUMNS

    annot_format_from_df = df_returned.annot_format.unique()
    assert len(annot_format_from_df) == 1
    annot_format_from_df = annot_format_from_df[0]

    if annot_format is not None:
        # test format from dataframe is specified format
        assert annot_format_from_df == annot_format
        annot_list = vak.common.annotation.from_df(df_returned)
    else:
        # if no annotation format specified, check that `annot_format` is `none`
        assert annot_format_from_df == vak.common.constants.NO_ANNOTATION_FORMAT
        annot_list = None

    if annot_file:
        assert all(
            [pathlib.Path(annot_path) == annot_file for annot_path in df_returned["annot_path"]]
        )

    if labelset:
        # test that filtering vocalizations by labelset worked
        assert annot_list is not None, "labelset specified but annot_list was None"
        for annot in annot_list:
            labels = annot.seq.labels
            labelset_from_labels = set(labels)
            # should be true that set of labels from each annotation is a subset of labelset
            assert labelset_from_labels.issubset(set(labelset))

    audio_paths_from_df = [pathlib.Path(audio_path) for audio_path in df_returned.audio_path]
    spect_paths_from_df = [pathlib.Path(spect_path) for spect_path in df_returned.spect_path]
    spect_file_names_from_df = [spect_path.name for spect_path in spect_paths_from_df]

    # which assertions to run depends on whether we made the dataframe
    # from audio files or spect files
    if audio_format:  # implies that --> we made the dataframe from audio files

        # test that all audio files came from data_dir
        for audio_path_from_df in audio_paths_from_df:
            assert audio_path_from_df.parent == data_dir

        # test that each audio file has a corresponding spect file in `spect_path` column
        expected_spect_files = [
            source_audio_file.name + spect_file_ext
            for source_audio_file in expected_audio_paths
        ]
        assert all(
            [
                expected_spect_file in spect_file_names_from_df
                for expected_spect_file in expected_spect_files
            ]
        )

        # if there are audio files we expect to **not** be in audio_path
        # -- because the associated annotations have labels not in labelset --
        # then test those files are **not** in spect_path
        if not_expected_audio_paths is not None:
            not_expected_spect_files = [
                source_audio_file.name + spect_file_ext
                for source_audio_file in not_expected_audio_paths
            ]
            assert all(
                [
                    not_expected_spect_file not in spect_file_names_from_df
                    for not_expected_spect_file in not_expected_spect_files
                ]
            )

        # test that all the generated spectrogram files are in spect_output_dir
        assert all(
            [
                spect_path.parents[0] == spect_output_dir
                for spect_path in spect_paths_from_df
            ]
        )

    elif spect_format:  # implies that --> we made the dataframe from spect files
        expected_spect_file_names = [
            spect_path.name for spect_path in expected_spect_paths
        ]

        assert all(
            [spect_file_name_from_df in expected_spect_file_names
             for spect_file_name_from_df in spect_file_names_from_df]
        )

        # test that **only** expected paths were in DataFrame
        if not_expected_spect_paths is not None:
            not_expected_spect_file_names = [
                spect_path.name for spect_path in not_expected_spect_paths
            ]
            assert all(
                [spect_file_name_from_df not in not_expected_spect_file_names
                 for spect_file_name_from_df in spect_file_names_from_df]
            )


def assert_spect_files_have_correct_keys(df_returned,
                                  spect_params):
    spect_paths_from_df = [pathlib.Path(spect_path) for spect_path in df_returned.spect_path]
    for spect_path in spect_paths_from_df:
        spect_dict = vak.common.files.spect.load(spect_path)
        for key_type in ['freqbins_key', 'timebins_key', 'spect_key']:
            if key_type in spect_params:
                key = spect_params[key_type]
            else:
                # if we didn't pass in this key type, don't check for it
                # this is for `audio_path` which is not strictly required currently
                continue
            assert key in spect_dict


@pytest.mark.parametrize(
    'data_dir, audio_format, spect_format, annot_format, labelset, annot_file',
    [
        (AUDIO_DIR_CBIN, "cbin", None, "notmat", True, None),
        (AUDIO_DIR_CBIN, "cbin", None, "notmat", False, None),
        (AUDIO_DIR_CBIN, "cbin", None, None, False, None),
        (SPECT_DIR_MAT, None, "mat", "yarden", True, ANNOT_FILE_YARDEN),
        (SPECT_DIR_MAT, None, "mat", "yarden", False, ANNOT_FILE_YARDEN),
        (SPECT_DIR_MAT, None, "mat", None, False, None),
    ]
)
def test_prep_spectrogram_dataset(
    data_dir,
    audio_format,
    spect_format,
    annot_format,
    labelset,
    annot_file,
    tmp_path,
    default_spect_params,
    specific_audio_list,
    specific_spect_list,
    specific_labelset,
):
    """Test that ``vak.prep.spectrogram_dataset.prep.prep_spectrogram_dataset`` works
    when we point it at directory of .cbin audio files
    and  **do not** specify an annotation format"""
    if labelset:
        labelset = specific_labelset(annot_format)
    else:
        labelset = None

    dataset_df = vak.prep.spectrogram_dataset.prep.prep_spectrogram_dataset(
        data_dir=data_dir,
        annot_format=annot_format,
        labelset=labelset,
        annot_file=annot_file,
        audio_format=audio_format,
        spect_output_dir=tmp_path,
        spect_format=spect_format,
        spect_params=default_spect_params,
    )

    if labelset and audio_format:
        expected_audio_paths = specific_audio_list(audio_format, "all_labels_in_labelset")
        not_expected_audio_paths = specific_audio_list(audio_format, "labels_not_in_labelset")
        expected_spect_paths = None
        not_expected_spect_paths = None
    elif labelset is None and audio_format:
        expected_audio_paths = specific_audio_list(audio_format)
        not_expected_audio_paths = None
        expected_spect_paths = None
        not_expected_spect_paths = None
    elif labelset and spect_format:
        expected_audio_paths = None
        not_expected_audio_paths = None
        expected_spect_paths = specific_spect_list(
            spect_format, "all_labels_in_labelset"
        )
        not_expected_spect_paths = specific_spect_list(
            spect_format, "labels_not_in_labelset"
        )
    elif labelset is None and spect_format:
        expected_audio_paths = None
        not_expected_audio_paths = None
        expected_spect_paths = specific_spect_list(spect_format)
        not_expected_spect_paths = None

    assert_returned_dataframe_matches_expected(
        dataset_df,
        data_dir=data_dir,
        annot_format=annot_format,
        labelset=labelset,
        audio_format=audio_format,
        spect_format=spect_format,
        spect_output_dir=tmp_path,
        expected_audio_paths=expected_audio_paths,
        not_expected_audio_paths=not_expected_audio_paths,
        expected_spect_paths=expected_spect_paths,
        not_expected_spect_paths=not_expected_spect_paths,
        annot_file=annot_file,
    )

    assert_spect_files_have_correct_keys(dataset_df, default_spect_params)



"""Functions for making a dataset of units from sequences,
as used to train dimensionality reduction models."""

from __future__ import annotations

import logging
import os
import pathlib

import attrs
import crowsetta
import dask
import dask.delayed
import numpy as np
import numpy.typing as npt
import pandas as pd
from dask.diagnostics import ProgressBar

from ...common import annotation, constants
from ...common.converters import expanded_user_path, labelset_to_set
from ..spectrogram_dataset.audio_helper import files_from_dir
from ..spectrogram_dataset.spect import spectrogram

logger = logging.getLogger(__name__)


@attrs.define
class Segment:
    """Dataclass that represents a segment
    from segmented audio or spectrogram.

    The attributes are metadata used to track
    the origin of this segment in a dataset
    of such segments.

    The dataset including metadata is saved as a csv file
    where these attributes become the columns.
    """

    data: npt.NDArray
    samplerate: int
    onset_s: float
    offset_s: float
    label: str
    sample_dur: float
    segment_dur: float
    audio_path: str
    annot_path: str


@dask.delayed
def get_segment_list(
    audio_path: str,
    annot: crowsetta.Annotation,
    audio_format: str,
    context_s: float = 0.005,
) -> list[Segment]:
    """Get a list of :class:`Segment` instances, given
    the path to an audio file and an annotation that indicates
    where segments occur in that audio file.

    Function used by
    :func:`vak.prep.dimensionality_reduction.unit_dataset.prep_unit_dataset`.

    Parameters
    ----------
    audio_path : str
        Path to an audio file.
    annot : crowsetta.Annotation
        Annotation for audio file.
    audio_format : str
        String representing audio file format, e.g. 'wav'.
    context_s : float
        Number of seconds of "context" around unit to
        add, i.e., time before and after the onset
        and offset respectively. Default is 0.005s,
        5 milliseconds.

    Returns
    -------
    segments : list
        A :class:`list` of :class:`Segment` instances.
    """
    data, samplerate = constants.AUDIO_FORMAT_FUNC_MAP[audio_format](
        audio_path
    )
    sample_dur = 1.0 / samplerate

    segments = []
    for onset_s, offset_s, label in zip(
        annot.seq.onsets_s, annot.seq.offsets_s, annot.seq.labels
    ):
        onset_s -= context_s
        offset_s += context_s
        onset_ind = int(np.floor(onset_s * samplerate))
        offset_ind = int(np.ceil(offset_s * samplerate))
        segment_data = data[onset_ind : offset_ind + 1]  # noqa: E203
        segment_dur = segment_data.shape[-1] * sample_dur
        segment = Segment(
            segment_data,
            samplerate,
            onset_s,
            offset_s,
            label,
            sample_dur,
            segment_dur,
            audio_path,
            annot.annot_path,
        )
        segments.append(segment)

    return segments


def spectrogram_from_segment(
    segment: Segment, spect_params: dict
) -> npt.NDArray:
    """Compute a spectrogram given a :class:`Segment` instance.

    Parameters
    ----------
    segment : Segment
    spect_params : dict

    Returns
    -------
    spect : numpy.ndarray
    """
    data, samplerate = np.array(segment.data), segment.samplerate
    s, _, _ = spectrogram(
        data,
        samplerate,
        spect_params.fft_size,
        spect_params.step_size,
        spect_params.thresh,
        spect_params.transform_type,
        spect_params.freq_cutoffs,
    )
    return s


@attrs.define
class SpectToSave:
    """A spectrogram to be saved.

    Used by :func:`save_spect`.
    """

    spect: npt.NDArray
    ind: int
    audio_path: str


def save_spect(
    spect_to_save: SpectToSave, output_dir: str | pathlib.Path
) -> str:
    """Save a spectrogram array to an npy file.

    The filename is build from the attributes of ``spect_to_save``,
    saved in output dir, and the full path is returned as a string.

    Parameters
    ----------
    spect_to_save : SpectToSave
    output_dir : str, pathlib.Path

    Returns
    -------
    npy_path : str
        Path to npy file containing spectrogram inside ``output_dir``
    """
    basename = (
        os.path.basename(spect_to_save.audio_path)
        + f"-segment-{spect_to_save.ind}"
    )
    npy_path = os.path.join(
        os.path.normpath(output_dir), basename + ".spect.npy"
    )
    np.save(npy_path, spect_to_save.spect)
    return npy_path


def abspath(a_path):
    """Convert a path to an absolute path"""
    if isinstance(a_path, str) or isinstance(a_path, pathlib.Path):
        return str(pathlib.Path(a_path).absolute())
    elif np.isnan(a_path):
        return a_path


# ---- make spectrograms + records for dataframe -----------------------------------------------------------------------
@dask.delayed
def make_spect_return_record(
    segment: Segment, ind: int, spect_params: dict, output_dir: pathlib.Path
) -> tuple:
    """Helper function that enables parallelized creation of "records",
    i.e. rows for dataframe, from .
    Accepts a two-element tuple containing (1) a dictionary that represents a spectrogram
    and (2) annotation for that file"""

    spect = spectrogram_from_segment(segment, spect_params)
    n_timebins = spect.shape[-1]

    spect_to_save = SpectToSave(spect, ind, segment.audio_path)
    spect_path = save_spect(spect_to_save, output_dir)
    record = tuple(
        [
            abspath(spect_path),
            abspath(segment.audio_path),
            abspath(segment.annot_path),
            segment.onset_s,
            segment.offset_s,
            segment.label,
            segment.samplerate,
            segment.sample_dur,
            segment.segment_dur,
        ]
    )

    return record, n_timebins


@dask.delayed
def pad_spectrogram(record: tuple, pad_length: float) -> None:
    """Pads a spectrogram to a specified length on the left and right sides.
    Spectrogram is saved again after padding.

    Parameters
    ----------
    record : tuple
    pad_length : int
    """
    spect_path = record[0]  # 'spect_path'
    spect = np.load(spect_path)

    excess_needed = pad_length - spect.shape[-1]
    pad_left = np.floor(float(excess_needed) / 2).astype("int")
    pad_right = np.ceil(float(excess_needed) / 2).astype("int")
    spect_padded = np.pad(
        spect, [(0, 0), (pad_left, pad_right)], "constant", constant_values=0
    )
    np.save(spect_path, spect_padded)
    return spect_padded.shape


# constant, used for names of columns in DataFrame below
DF_COLUMNS = [
    "spect_path",
    "audio_path",
    "annot_path",
    "onset_s",
    "offset_s",
    "label",
    "samplerate",
    "sample_dur",
    "duration",
]


def prep_unit_dataset(
    audio_format: str,
    output_dir: str,
    spect_params: dict,
    data_dir: list | None = None,
    annot_format: str | None = None,
    annot_file: str | pathlib.Path | None = None,
    labelset: set | None = None,
    context_s: float = 0.005,
) -> pd.DataFrame:
    """Prepare a dataset of units from sequences,
    e.g., all syllables segmented out of a dataset of birdsong.

    Parameters
    ----------
    audio_format
    output_dir
    spect_params
    data_dir
    annot_format
    annot_file
    labelset
    context_s

    Returns
    -------
    unit_df : pandas.DataFrame
        A DataFrame representing all the units in the dataset.
    shape: tuple
        A tuple representing the shape of all spectograms in the dataset.
        The spectrograms of all units are padded so that they are all
        as wide as the widest unit (i.e, the one with the longest duration).
    """
    # pre-conditions ---------------------------------------------------------------------------------------------------
    if audio_format not in constants.VALID_AUDIO_FORMATS:
        raise ValueError(
            f"audio format must be one of '{constants.VALID_AUDIO_FORMATS}'; "
            f"format '{audio_format}' not recognized."
        )

    if labelset is not None:
        labelset = labelset_to_set(labelset)

    data_dir = expanded_user_path(data_dir)
    if not data_dir.is_dir():
        raise NotADirectoryError(f"data_dir not found: {data_dir}")

    audio_files = files_from_dir(data_dir, audio_format)

    if annot_format is not None:
        if annot_file is None:
            annot_files = annotation.files_from_dir(
                annot_dir=data_dir, annot_format=annot_format
            )
            scribe = crowsetta.Transcriber(format=annot_format)
            annot_list = [
                scribe.from_file(annot_file).to_annot()
                for annot_file in annot_files
            ]
        else:
            scribe = crowsetta.Transcriber(format=annot_format)
            annot_list = scribe.from_file(annot_file).to_annot()
        if isinstance(annot_list, crowsetta.Annotation):
            # if e.g. only one annotated audio file in directory, wrap in a list to make iterable
            # fixes https://github.com/NickleDave/vak/issues/467
            annot_list = [annot_list]
    else:  # if annot_format not specified
        annot_list = None

    if annot_list:
        audio_annot_map = annotation.map_annotated_to_annot(
            audio_files, annot_list, annot_format
        )
    else:
        # no annotation, so map spectrogram files to None
        audio_annot_map = dict(
            (audio_path, None) for audio_path in audio_files
        )

    # use labelset, if supplied, with annotations, if any, to filter;
    if (
        labelset and annot_list
    ):  # then remove annotations with labels not in labelset
        for audio_file, annot in list(audio_annot_map.items()):
            # loop in a verbose way (i.e. not a comprehension)
            # so we can give user warning when we skip files
            annot_labelset = set(annot.seq.labels)
            # below, set(labels_mapping) is a set of that dict's keys
            if not annot_labelset.issubset(set(labelset)):
                # because there's some label in labels that's not in labelset
                audio_annot_map.pop(audio_file)
                extra_labels = annot_labelset - labelset
                logger.info(
                    f"Found labels, {extra_labels}, in {pathlib.Path(audio_file).name}, "
                    "that are not in labels_mapping. Skipping file.",
                )

    segments = []
    for audio_path, annot in audio_annot_map.items():
        segment_list = dask.delayed(get_segment_list)(
            audio_path, annot, audio_format, context_s
        )
        segments.append(segment_list)

    logger.info(
        "Loading audio for all segments in all files",
    )
    with ProgressBar():
        segments: list[list[Segment]] = dask.compute(*segments)
    segments: list[Segment] = [
        segment for segment_list in segments for segment in segment_list
    ]

    # ---- make and save all spectrograms *before* padding
    # This is a design choice to avoid keeping all the spectrograms in memory
    # but since we want to pad all spectrograms to be the same width,
    # it requires us to go back, load each one, and pad it.
    # Might be worth looking at how often typical dataset sizes in memory and whether this is really necessary.
    records_n_timebins_tuples = []
    for ind, segment in enumerate(segments):
        records_n_timebins_tuple = make_spect_return_record(
            segment, ind, spect_params, output_dir
        )
        records_n_timebins_tuples.append(records_n_timebins_tuple)
    with ProgressBar():
        records_n_timebins_tuples: list[tuple[tuple, int]] = dask.compute(
            *records_n_timebins_tuples
        )

    records, n_timebins_list = [], []
    for records_n_timebins_tuple in records_n_timebins_tuples:
        record, n_timebins = records_n_timebins_tuple
        records.append(record)
        n_timebins_list.append(n_timebins)

    pad_length = max(n_timebins_list)

    padded = []
    for record in records:
        padded.append(pad_spectrogram(record, pad_length))
    with ProgressBar():
        shapes: list[tuple[int, int]] = dask.compute(*padded)

    shape = set(shapes)
    assert (
        len(shape) == 1
    ), f"Did not find a single unique shape for all spectrograms. Instead found: {shape}"
    shape = shape.pop()

    unit_df = pd.DataFrame.from_records(records, columns=DF_COLUMNS)

    return unit_df, shape
